# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/frontend.proto
# plugin: python-betterproto
# This file has been @generated
import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Dict,
    List,
    Optional,
)

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import (
    compiler as _compiler__,
    general_messages as _general_messages__,
    qm_api as _qm_api__,
    qm_manager as _qm_manager__,
    qua as _qua__,
    qua_config as _qua_config__,
    results_analyser as _results_analyser__,
)


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class IsJobAcquiringDataResponseAcquiringStatus(betterproto.Enum):
    ACQUIRE_STOPPED = 0
    NO_DATA_TO_ACQUIRE = 1
    HAS_DATA_TO_ACQUIRE = 2


@dataclass(eq=False, repr=False)
class QueuePosition(betterproto.Message):
    end: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        1, group="position"
    )
    start: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        2, group="position"
    )


@dataclass(eq=False, repr=False)
class AddToQueueRequest(betterproto.Message):
    quantum_machine_id: str = betterproto.string_field(1)
    high_level_program: "_qua__.QuaProgram" = betterproto.message_field(2)
    queue_position: "QueuePosition" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class AddToQueueResponse(betterproto.Message):
    ok: bool = betterproto.bool_field(1)
    job_id: str = betterproto.string_field(2)
    messages: List["_compiler__.CompilerMessage"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class WaveformOverride(betterproto.Message):
    samples: List[float] = betterproto.double_field(1)


@dataclass(eq=False, repr=False)
class ExecutionOverrides(betterproto.Message):
    waveforms: Dict[str, "WaveformOverride"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class AddCompiledToQueueRequest(betterproto.Message):
    quantum_machine_id: str = betterproto.string_field(1)
    program_id: str = betterproto.string_field(2)
    queue_position: "QueuePosition" = betterproto.message_field(3)
    execution_overrides: "ExecutionOverrides" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class AddCompiledToQueueResponse(betterproto.Message):
    ok: bool = betterproto.bool_field(1)
    job_id: str = betterproto.string_field(2)
    errors: List["_general_messages__.ErrorMessage"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class CompileRequest(betterproto.Message):
    quantum_machine_id: str = betterproto.string_field(1)
    high_level_program: "_qua__.QuaProgram" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class CompileResponse(betterproto.Message):
    ok: bool = betterproto.bool_field(1)
    program_id: str = betterproto.string_field(2)
    messages: List["_compiler__.CompilerMessage"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class GetJobExecutionStatusRequest(betterproto.Message):
    quantum_machine_id: str = betterproto.string_field(1)
    job_id: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class GetJobExecutionStatusResponse(betterproto.Message):
    status: "JobExecutionStatus" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class JobExecutionStatus(betterproto.Message):
    unknown: "JobExecutionStatusUnknown" = betterproto.message_field(1, group="status")
    pending: "JobExecutionStatusPending" = betterproto.message_field(2, group="status")
    running: "JobExecutionStatusRunning" = betterproto.message_field(3, group="status")
    completed: "JobExecutionStatusCompleted" = betterproto.message_field(
        4, group="status"
    )
    canceled: "JobExecutionStatusCanceled" = betterproto.message_field(
        5, group="status"
    )
    loading: "JobExecutionStatusLoading" = betterproto.message_field(6, group="status")
    error: "JobExecutionStatusError" = betterproto.message_field(7, group="status")
    processing: "JobExecutionStatusProcessing" = betterproto.message_field(
        8, group="status"
    )


@dataclass(eq=False, repr=False)
class JobExecutionStatusUnknown(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class JobExecutionStatusPending(betterproto.Message):
    position_in_queue: int = betterproto.int32_field(1)
    time_added: datetime = betterproto.message_field(2)
    added_by: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class JobExecutionStatusRunning(betterproto.Message):
    time_added: datetime = betterproto.message_field(1)
    added_by: str = betterproto.string_field(2)
    time_started: datetime = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class JobExecutionStatusCompleted(betterproto.Message):
    time_added: datetime = betterproto.message_field(1)
    added_by: str = betterproto.string_field(2)
    time_started: datetime = betterproto.message_field(3)
    time_completed: datetime = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class JobExecutionStatusCanceled(betterproto.Message):
    time_added: datetime = betterproto.message_field(1)
    added_by: str = betterproto.string_field(2)
    time_canceled: datetime = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class JobExecutionStatusLoading(betterproto.Message):
    time_added: datetime = betterproto.message_field(1)
    added_by: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class JobExecutionStatusError(betterproto.Message):
    error_messages: "betterproto_lib_google_protobuf.ListValue" = (
        betterproto.message_field(1)
    )
    added_by: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class JobExecutionStatusProcessing(betterproto.Message):
    time_added: datetime = betterproto.message_field(1)
    added_by: str = betterproto.string_field(2)
    time_started: datetime = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class GetPendingJobsResponse(betterproto.Message):
    pending_jobs: Dict[str, "JobExecutionStatusPending"] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )


@dataclass(eq=False, repr=False)
class JobQueryParams(betterproto.Message):
    quantum_machine_id: str = betterproto.string_field(1)
    job_id: "QueryValueMatcher" = betterproto.message_field(2)
    user_id: "QueryValueMatcher" = betterproto.message_field(3)
    position: Optional[int] = betterproto.message_field(
        4, wraps=betterproto.TYPE_UINT32
    )


@dataclass(eq=False, repr=False)
class QueryValueMatcher(betterproto.Message):
    any: bool = betterproto.bool_field(1, group="valueType")
    value: str = betterproto.string_field(2, group="valueType")


@dataclass(eq=False, repr=False)
class RemovePendingJobsResponse(betterproto.Message):
    numbers_of_jobs_removed: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class SimulationRequest(betterproto.Message):
    config: "_qua_config__.QuaConfig" = betterproto.message_field(1)
    high_level_program: "_qua__.QuaProgram" = betterproto.message_field(2)
    simulate: "ExecutionRequestSimulate" = betterproto.message_field(3)
    """The duration to simulate and provide results for"""

    controller_connections: List["InterOpxConnection"] = betterproto.message_field(6)


@dataclass(eq=False, repr=False)
class InterOpxAddress(betterproto.Message):
    controller: str = betterproto.string_field(1)
    left: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class InterOpxTarget(betterproto.Message):
    direct: "InterOpxAddress" = betterproto.message_field(1, group="target")


@dataclass(eq=False, repr=False)
class InterOpxChannel(betterproto.Message):
    controller: str = betterproto.string_field(1)
    channel_number: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class InterOpxConnection(betterproto.Message):
    source: "InterOpxAddress" = betterproto.message_field(1)
    target: "InterOpxTarget" = betterproto.message_field(2)
    address_to_address: "InterOpxConnectionAddressToAddress" = (
        betterproto.message_field(3, group="connectionType")
    )
    channel_to_channel: "InterOpxConnectionChannelToChannel" = (
        betterproto.message_field(4, group="connectionType")
    )

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("source"):
            warnings.warn("InterOpxConnection.source is deprecated", DeprecationWarning)
        if self.is_set("target"):
            warnings.warn("InterOpxConnection.target is deprecated", DeprecationWarning)


@dataclass(eq=False, repr=False)
class InterOpxConnectionAddressToAddress(betterproto.Message):
    source: "InterOpxAddress" = betterproto.message_field(1)
    target: "InterOpxAddress" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class InterOpxConnectionChannelToChannel(betterproto.Message):
    source: "InterOpxChannel" = betterproto.message_field(1)
    target: "InterOpxChannel" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class SimulationResponse(betterproto.Message):
    success: bool = betterproto.bool_field(1)
    job_id: str = betterproto.string_field(2)
    config_validation_errors: List[
        "_qm_manager__.ConfigValidationMessage"
    ] = betterproto.message_field(3)
    physical_validation_errors: List[
        "_qm_manager__.PhysicalValidationMessage"
    ] = betterproto.message_field(4)
    messages: List["_compiler__.CompilerMessage"] = betterproto.message_field(5)
    simulated: "SimulatedResponsePart" = betterproto.message_field(6)


@dataclass(eq=False, repr=False)
class IsJobRunningRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class IsJobRunningResponse(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    is_running: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class IsJobAcquiringDataRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class IsJobAcquiringDataResponse(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    acquiring_status: "IsJobAcquiringDataResponseAcquiringStatus" = (
        betterproto.enum_field(2)
    )


@dataclass(eq=False, repr=False)
class QmDataRequest(betterproto.Message):
    io_value_request: List["IoValueRequest"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QmDataResponse(betterproto.Message):
    io_value_response: List[
        "QmDataResponseIoValueResponse"
    ] = betterproto.message_field(1)
    success: bool = betterproto.bool_field(2)
    errors: List["_general_messages__.ErrorMessage"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class QmDataResponseIoValueResponse(betterproto.Message):
    request: "IoValueRequest" = betterproto.message_field(1)
    values: "_compiler__.QuaValues" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class IoValueRequest(betterproto.Message):
    pulser_number: int = betterproto.uint32_field(1)
    io_number: int = betterproto.uint32_field(2)
    job_id: str = betterproto.string_field(3, group="identifier_one_of")
    quantum_machine_id: str = betterproto.string_field(4, group="identifier_one_of")

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("pulser_number"):
            warnings.warn(
                "IoValueRequest.pulser_number is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class ResumeRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    pulsers_bitmap: int = betterproto.uint32_field(2)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("pulsers_bitmap"):
            warnings.warn(
                "ResumeRequest.pulsers_bitmap is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class ResumeResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PausedStatusRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class PausedStatusResponse(betterproto.Message):
    pulsers_bitmap: int = betterproto.uint32_field(1)
    ok: bool = betterproto.bool_field(2)
    is_paused: bool = betterproto.bool_field(3)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("pulsers_bitmap"):
            warnings.warn(
                "PausedStatusResponse.pulsers_bitmap is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class PullResultRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class PullAnalysedResultsRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    metadata: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class PullResultResponse(betterproto.Message):
    data: bytes = betterproto.bytes_field(1)
    offset: int = betterproto.uint32_field(2)
    bytes_left: int = betterproto.uint32_field(3)
    controller_name: str = betterproto.string_field(4)
    application_number: int = betterproto.uint32_field(5)
    implementation_name: str = betterproto.string_field(6)


@dataclass(eq=False, repr=False)
class HealthCheckResponse(betterproto.Message):
    message: List[str] = betterproto.string_field(1)
    ok: bool = betterproto.bool_field(2)
    error_messages: List[str] = betterproto.string_field(3)
    warning_messages: List[str] = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class ExecutionRequest(betterproto.Message):
    """Execution request"""

    high_level_program: "_qua__.QuaProgram" = betterproto.message_field(
        1, group="program_one_of"
    )
    low_level_program: bytes = betterproto.bytes_field(2, group="program_one_of")
    stream_duration_limit: int = betterproto.uint32_field(3)
    """
    Stream limitation, time in milliseconds and data in records (4ns worth of raw data
     or a single icp/seq record).
    """

    stream_data_limit: int = betterproto.uint32_field(4)
    force_execution: bool = betterproto.bool_field(5)
    """
    Force the execution. If another job is currently running, terminate it.
    """

    simulate: "ExecutionRequestSimulate" = betterproto.message_field(6)
    """The duration to simulate and provide results for"""

    dry_run: bool = betterproto.bool_field(7)
    quantum_machine_id: str = betterproto.string_field(8)
    """The id of the qm that will execute the program"""

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("low_level_program"):
            warnings.warn(
                "ExecutionRequest.low_level_program is deprecated", DeprecationWarning
            )


@dataclass(eq=False, repr=False)
class ExecutionRequestSimulate(betterproto.Message):
    duration: int = betterproto.uint32_field(1)
    simulate_analog_outputs: bool = betterproto.bool_field(2)
    simulate_digital_outputs: bool = betterproto.bool_field(3)
    include_analog_samples: bool = betterproto.bool_field(4)
    include_digital_samples: bool = betterproto.bool_field(5)
    include_analog_waveforms: bool = betterproto.bool_field(6)
    include_digital_waveforms: bool = betterproto.bool_field(7)
    simulation_interface: "ExecutionRequestSimulateSimulationInterface" = (
        betterproto.message_field(8)
    )
    extra_processing_timeout_ms: int = betterproto.int32_field(9)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("simulate_analog_outputs"):
            warnings.warn(
                "ExecutionRequestSimulate.simulate_analog_outputs is deprecated",
                DeprecationWarning,
            )
        if self.is_set("simulate_digital_outputs"):
            warnings.warn(
                "ExecutionRequestSimulate.simulate_digital_outputs is deprecated",
                DeprecationWarning,
            )
        if self.is_set("include_analog_samples"):
            warnings.warn(
                "ExecutionRequestSimulate.include_analog_samples is deprecated",
                DeprecationWarning,
            )
        if self.is_set("include_digital_samples"):
            warnings.warn(
                "ExecutionRequestSimulate.include_digital_samples is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class ExecutionRequestSimulateSimulationInterface(betterproto.Message):
    none: "ExecutionRequestSimulateSimulationInterfaceNone" = betterproto.message_field(
        1, group="interface_oneof"
    )
    loopback: "ExecutionRequestSimulateSimulationInterfaceLoopback" = (
        betterproto.message_field(2, group="interface_oneof")
    )
    qsim: "ExecutionRequestSimulateSimulationInterfaceQSim" = betterproto.message_field(
        3, group="interface_oneof"
    )
    raw: "ExecutionRequestSimulateSimulationInterfaceRawInterface" = (
        betterproto.message_field(4, group="interface_oneof")
    )
    quantum_simulator: "ExecutionRequestSimulateSimulationInterfaceQuantumSimulatorInterface" = betterproto.message_field(
        5, group="interface_oneof"
    )


@dataclass(eq=False, repr=False)
class ExecutionRequestSimulateSimulationInterfaceNone(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ExecutionRequestSimulateSimulationInterfaceLoopback(betterproto.Message):
    connections: List[
        "ExecutionRequestSimulateSimulationInterfaceLoopbackConnections"
    ] = betterproto.message_field(2)
    latency: int = betterproto.uint32_field(3)
    noise_power: float = betterproto.double_field(4)


@dataclass(eq=False, repr=False)
class ExecutionRequestSimulateSimulationInterfaceLoopbackConnections(
    betterproto.Message
):
    from_controller: str = betterproto.string_field(1)
    from_fem: int = betterproto.int32_field(5)
    from_port: int = betterproto.int32_field(2)
    to_controller: str = betterproto.string_field(3)
    to_fem: int = betterproto.int32_field(6)
    to_port: int = betterproto.int32_field(4)


@dataclass(eq=False, repr=False)
class ExecutionRequestSimulateSimulationInterfaceQSim(betterproto.Message):
    """TODO remove old quantum sim"""

    target: str = betterproto.string_field(1)
    job_id: str = betterproto.string_field(2)
    analog_out_connections: List[
        "ExecutionRequestSimulateSimulationInterfaceQSimConnections"
    ] = betterproto.message_field(3)
    digital_out_connections: List[
        "ExecutionRequestSimulateSimulationInterfaceQSimConnections"
    ] = betterproto.message_field(4)
    analog_in_connections: List[
        "ExecutionRequestSimulateSimulationInterfaceQSimConnections"
    ] = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class ExecutionRequestSimulateSimulationInterfaceQSimConnections(betterproto.Message):
    from_controller: str = betterproto.string_field(1)
    from_port: int = betterproto.int32_field(2)
    to_q_sim: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class ExecutionRequestSimulateSimulationInterfaceQuantumSimulatorInterface(
    betterproto.Message
):
    physical_config: bytes = betterproto.bytes_field(1)


@dataclass(eq=False, repr=False)
class ExecutionRequestSimulateSimulationInterfaceRawInterface(betterproto.Message):
    connections: List[
        "ExecutionRequestSimulateSimulationInterfaceRawInterfaceConnections"
    ] = betterproto.message_field(1)
    noise_power: float = betterproto.double_field(2)


@dataclass(eq=False, repr=False)
class ExecutionRequestSimulateSimulationInterfaceRawInterfaceConnections(
    betterproto.Message
):
    from_controller: str = betterproto.string_field(1)
    from_fem: int = betterproto.int32_field(4)
    from_port: int = betterproto.int32_field(2)
    to_samples: List[float] = betterproto.double_field(3)


@dataclass(eq=False, repr=False)
class ExecutionResponse(betterproto.Message):
    ok: bool = betterproto.bool_field(1)
    job_id: str = betterproto.string_field(2)
    messages: List["_compiler__.CompilerMessage"] = betterproto.message_field(3)
    metadata: str = betterproto.string_field(4)
    simulated: "SimulatedResponsePart" = betterproto.message_field(6)
    config: "_qua_config__.QuaConfig" = betterproto.message_field(7)
    """The config used for execution (only filled for high level execution)"""


@dataclass(eq=False, repr=False)
class SimulatedResponsePart(betterproto.Message):
    analog_outputs: "betterproto_lib_google_protobuf.Struct" = (
        betterproto.message_field(1)
    )
    digital_outputs: "betterproto_lib_google_protobuf.Struct" = (
        betterproto.message_field(2)
    )
    waveform_report: "betterproto_lib_google_protobuf.Struct" = (
        betterproto.message_field(4)
    )
    errors: List[str] = betterproto.string_field(3)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("analog_outputs"):
            warnings.warn(
                "SimulatedResponsePart.analog_outputs is deprecated", DeprecationWarning
            )
        if self.is_set("digital_outputs"):
            warnings.warn(
                "SimulatedResponsePart.digital_outputs is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class ResetDataProcessingRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ResetDataProcessingResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class HaltRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class HaltResponse(betterproto.Message):
    ok: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class PeekRequest(betterproto.Message):
    address: int = betterproto.uint32_field(1)
    controller_id: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class PeekResponse(betterproto.Message):
    value: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class PokeRequest(betterproto.Message):
    address: int = betterproto.uint32_field(1)
    value: int = betterproto.uint32_field(2)
    controller_id: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class PokeResponse(betterproto.Message):
    ok: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class GetSimulatedQuantumStateRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetSimulatedQuantumStateResponse(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    ok: bool = betterproto.bool_field(2)
    state: "DensityMatrix" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class DensityMatrix(betterproto.Message):
    time_stamp: int = betterproto.uint64_field(1)
    data: List["ComplexNumber"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class ComplexNumber(betterproto.Message):
    re: float = betterproto.double_field(1)
    im: float = betterproto.double_field(2)


@dataclass(eq=False, repr=False)
class PerformHalDebugCommandRequest(betterproto.Message):
    controller_name: str = betterproto.string_field(1)
    command: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class PerformHalDebugCommandResponse(betterproto.Message):
    success: bool = betterproto.bool_field(1)
    response: str = betterproto.string_field(2)


class FrontendStub(betterproto.ServiceStub):
    async def get_version(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "betterproto_lib_google_protobuf.StringValue":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/GetVersion",
            betterproto_lib_google_protobuf_empty,
            betterproto_lib_google_protobuf.StringValue,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def health_check(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "HealthCheckResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/HealthCheck",
            betterproto_lib_google_protobuf_empty,
            HealthCheckResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def execute(
        self,
        execution_request: "ExecutionRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "ExecutionResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/Execute",
            execution_request,
            ExecutionResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def add_to_queue(
        self,
        add_to_queue_request: "AddToQueueRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "AddToQueueResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/AddToQueue",
            add_to_queue_request,
            AddToQueueResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def compile(
        self,
        compile_request: "CompileRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "CompileResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/Compile",
            compile_request,
            CompileResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def add_compiled_to_queue(
        self,
        add_compiled_to_queue_request: "AddCompiledToQueueRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "AddCompiledToQueueResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/AddCompiledToQueue",
            add_compiled_to_queue_request,
            AddCompiledToQueueResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def remove_pending_jobs(
        self,
        job_query_params: "JobQueryParams",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "RemovePendingJobsResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/RemovePendingJobs",
            job_query_params,
            RemovePendingJobsResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_pending_jobs(
        self,
        job_query_params: "JobQueryParams",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "GetPendingJobsResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/GetPendingJobs",
            job_query_params,
            GetPendingJobsResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_job_execution_status(
        self,
        get_job_execution_status_request: "GetJobExecutionStatusRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "GetJobExecutionStatusResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/GetJobExecutionStatus",
            get_job_execution_status_request,
            GetJobExecutionStatusResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_job_status_updates(
        self,
        get_job_execution_status_request: "GetJobExecutionStatusRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> AsyncIterator[GetJobExecutionStatusResponse]:
        async for response in self._unary_stream(
            "/qm.grpc.frontend.Frontend/GetJobStatusUpdates",
            get_job_execution_status_request,
            GetJobExecutionStatusResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response

    async def paused_status(
        self,
        paused_status_request: "PausedStatusRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "PausedStatusResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/PausedStatus",
            paused_status_request,
            PausedStatusResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def resume(
        self,
        resume_request: "ResumeRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "ResumeResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/Resume",
            resume_request,
            ResumeResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def halt(
        self,
        halt_request: "HaltRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "HaltResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/Halt",
            halt_request,
            HaltResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def reset_data_processing(
        self,
        reset_data_processing_request: "ResetDataProcessingRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "ResetDataProcessingResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/ResetDataProcessing",
            reset_data_processing_request,
            ResetDataProcessingResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def validate_config(
        self,
        qua_config_qua_config: "_qua_config__.QuaConfig",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "_compiler__.ValidationResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/ValidateConfig",
            qua_config_qua_config,
            _compiler__.ValidationResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def init(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "betterproto_lib_google_protobuf.Empty":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/Init",
            betterproto_lib_google_protobuf_empty,
            betterproto_lib_google_protobuf.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def perform_qm_request(
        self,
        qm_api_high_qm_api_request: "_qm_api__.HighQmApiRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "_qm_api__.HighQmApiResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/PerformQmRequest",
            qm_api_high_qm_api_request,
            _qm_api__.HighQmApiResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def request_data(
        self,
        qm_data_request: "QmDataRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "QmDataResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/RequestData",
            qm_data_request,
            QmDataResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def pull_simulator_samples(
        self,
        results_analyser_pull_simulator_samples_request: "_results_analyser__.PullSimulatorSamplesRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> AsyncIterator[_results_analyser__.SimulatorSamplesResponse]:
        async for response in self._unary_stream(
            "/qm.grpc.frontend.Frontend/PullSimulatorSamples",
            results_analyser_pull_simulator_samples_request,
            _results_analyser__.SimulatorSamplesResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response

    async def open_quantum_machine(
        self,
        qm_manager_open_quantum_machine_request: "_qm_manager__.OpenQuantumMachineRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "_qm_manager__.OpenQuantumMachineResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/OpenQuantumMachine",
            qm_manager_open_quantum_machine_request,
            _qm_manager__.OpenQuantumMachineResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def close_quantum_machine(
        self,
        qm_manager_close_quantum_machine_request: "_qm_manager__.CloseQuantumMachineRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "_qm_manager__.CloseQuantumMachineResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/CloseQuantumMachine",
            qm_manager_close_quantum_machine_request,
            _qm_manager__.CloseQuantumMachineResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_quantum_machine(
        self,
        qm_manager_get_quantum_machine_request: "_qm_manager__.GetQuantumMachineRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "_qm_manager__.GetQuantumMachineResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/GetQuantumMachine",
            qm_manager_get_quantum_machine_request,
            _qm_manager__.GetQuantumMachineResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_running_job(
        self,
        qm_manager_get_running_job_request: "_qm_manager__.GetRunningJobRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "_qm_manager__.GetRunningJobResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/GetRunningJob",
            qm_manager_get_running_job_request,
            _qm_manager__.GetRunningJobResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def list_open_quantum_machines(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "_qm_manager__.ListOpenQuantumMachinesResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/ListOpenQuantumMachines",
            betterproto_lib_google_protobuf_empty,
            _qm_manager__.ListOpenQuantumMachinesResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def close_all_quantum_machines(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "_qm_manager__.CloseAllQuantumMachinesResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/CloseAllQuantumMachines",
            betterproto_lib_google_protobuf_empty,
            _qm_manager__.CloseAllQuantumMachinesResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_controllers(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "_qm_manager__.GetControllersResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/GetControllers",
            betterproto_lib_google_protobuf_empty,
            _qm_manager__.GetControllersResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def is_job_running(
        self,
        is_job_running_request: "IsJobRunningRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "IsJobRunningResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/IsJobRunning",
            is_job_running_request,
            IsJobRunningResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def is_job_acquiring_data(
        self,
        is_job_acquiring_data_request: "IsJobAcquiringDataRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "IsJobAcquiringDataResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/IsJobAcquiringData",
            is_job_acquiring_data_request,
            IsJobAcquiringDataResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def simulate(
        self,
        simulation_request: "SimulationRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "SimulationResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/Simulate",
            simulation_request,
            SimulationResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def clear_all_job_results(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "betterproto_lib_google_protobuf.Empty":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/ClearAllJobResults",
            betterproto_lib_google_protobuf_empty,
            betterproto_lib_google_protobuf.Empty,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def perform_hal_debug_command(
        self,
        perform_hal_debug_command_request: "PerformHalDebugCommandRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "PerformHalDebugCommandResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/PerformHalDebugCommand",
            perform_hal_debug_command_request,
            PerformHalDebugCommandResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def get_simulated_quantum_state(
        self,
        get_simulated_quantum_state_request: "GetSimulatedQuantumStateRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "GetSimulatedQuantumStateResponse":
        return await self._unary_unary(
            "/qm.grpc.frontend.Frontend/GetSimulatedQuantumState",
            get_simulated_quantum_state_request,
            GetSimulatedQuantumStateResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class FrontendBase(ServiceBase):
    async def get_version(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
    ) -> "betterproto_lib_google_protobuf.StringValue":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def health_check(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
    ) -> "HealthCheckResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def execute(
        self, execution_request: "ExecutionRequest"
    ) -> "ExecutionResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def add_to_queue(
        self, add_to_queue_request: "AddToQueueRequest"
    ) -> "AddToQueueResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def compile(self, compile_request: "CompileRequest") -> "CompileResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def add_compiled_to_queue(
        self, add_compiled_to_queue_request: "AddCompiledToQueueRequest"
    ) -> "AddCompiledToQueueResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def remove_pending_jobs(
        self, job_query_params: "JobQueryParams"
    ) -> "RemovePendingJobsResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_pending_jobs(
        self, job_query_params: "JobQueryParams"
    ) -> "GetPendingJobsResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_job_execution_status(
        self, get_job_execution_status_request: "GetJobExecutionStatusRequest"
    ) -> "GetJobExecutionStatusResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_job_status_updates(
        self, get_job_execution_status_request: "GetJobExecutionStatusRequest"
    ) -> AsyncIterator[GetJobExecutionStatusResponse]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
        yield GetJobExecutionStatusResponse()

    async def paused_status(
        self, paused_status_request: "PausedStatusRequest"
    ) -> "PausedStatusResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def resume(self, resume_request: "ResumeRequest") -> "ResumeResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def halt(self, halt_request: "HaltRequest") -> "HaltResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def reset_data_processing(
        self, reset_data_processing_request: "ResetDataProcessingRequest"
    ) -> "ResetDataProcessingResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def validate_config(
        self, qua_config_qua_config: "_qua_config__.QuaConfig"
    ) -> "_compiler__.ValidationResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def init(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
    ) -> "betterproto_lib_google_protobuf.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def perform_qm_request(
        self, qm_api_high_qm_api_request: "_qm_api__.HighQmApiRequest"
    ) -> "_qm_api__.HighQmApiResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def request_data(self, qm_data_request: "QmDataRequest") -> "QmDataResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def pull_simulator_samples(
        self,
        results_analyser_pull_simulator_samples_request: "_results_analyser__.PullSimulatorSamplesRequest",
    ) -> AsyncIterator[_results_analyser__.SimulatorSamplesResponse]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)
        yield _results_analyser__.SimulatorSamplesResponse()

    async def open_quantum_machine(
        self,
        qm_manager_open_quantum_machine_request: "_qm_manager__.OpenQuantumMachineRequest",
    ) -> "_qm_manager__.OpenQuantumMachineResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def close_quantum_machine(
        self,
        qm_manager_close_quantum_machine_request: "_qm_manager__.CloseQuantumMachineRequest",
    ) -> "_qm_manager__.CloseQuantumMachineResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_quantum_machine(
        self,
        qm_manager_get_quantum_machine_request: "_qm_manager__.GetQuantumMachineRequest",
    ) -> "_qm_manager__.GetQuantumMachineResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_running_job(
        self, qm_manager_get_running_job_request: "_qm_manager__.GetRunningJobRequest"
    ) -> "_qm_manager__.GetRunningJobResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def list_open_quantum_machines(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
    ) -> "_qm_manager__.ListOpenQuantumMachinesResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def close_all_quantum_machines(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
    ) -> "_qm_manager__.CloseAllQuantumMachinesResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_controllers(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
    ) -> "_qm_manager__.GetControllersResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def is_job_running(
        self, is_job_running_request: "IsJobRunningRequest"
    ) -> "IsJobRunningResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def is_job_acquiring_data(
        self, is_job_acquiring_data_request: "IsJobAcquiringDataRequest"
    ) -> "IsJobAcquiringDataResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def simulate(
        self, simulation_request: "SimulationRequest"
    ) -> "SimulationResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def clear_all_job_results(
        self,
        betterproto_lib_google_protobuf_empty: "betterproto_lib_google_protobuf.Empty",
    ) -> "betterproto_lib_google_protobuf.Empty":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def perform_hal_debug_command(
        self, perform_hal_debug_command_request: "PerformHalDebugCommandRequest"
    ) -> "PerformHalDebugCommandResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def get_simulated_quantum_state(
        self, get_simulated_quantum_state_request: "GetSimulatedQuantumStateRequest"
    ) -> "GetSimulatedQuantumStateResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_get_version(
        self,
        stream: "grpclib.server.Stream[betterproto_lib_google_protobuf.Empty, betterproto_lib_google_protobuf.StringValue]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_version(request)
        await stream.send_message(response)

    async def __rpc_health_check(
        self,
        stream: "grpclib.server.Stream[betterproto_lib_google_protobuf.Empty, HealthCheckResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.health_check(request)
        await stream.send_message(response)

    async def __rpc_execute(
        self, stream: "grpclib.server.Stream[ExecutionRequest, ExecutionResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.execute(request)
        await stream.send_message(response)

    async def __rpc_add_to_queue(
        self, stream: "grpclib.server.Stream[AddToQueueRequest, AddToQueueResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.add_to_queue(request)
        await stream.send_message(response)

    async def __rpc_compile(
        self, stream: "grpclib.server.Stream[CompileRequest, CompileResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.compile(request)
        await stream.send_message(response)

    async def __rpc_add_compiled_to_queue(
        self,
        stream: "grpclib.server.Stream[AddCompiledToQueueRequest, AddCompiledToQueueResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.add_compiled_to_queue(request)
        await stream.send_message(response)

    async def __rpc_remove_pending_jobs(
        self, stream: "grpclib.server.Stream[JobQueryParams, RemovePendingJobsResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.remove_pending_jobs(request)
        await stream.send_message(response)

    async def __rpc_get_pending_jobs(
        self, stream: "grpclib.server.Stream[JobQueryParams, GetPendingJobsResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_pending_jobs(request)
        await stream.send_message(response)

    async def __rpc_get_job_execution_status(
        self,
        stream: "grpclib.server.Stream[GetJobExecutionStatusRequest, GetJobExecutionStatusResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_job_execution_status(request)
        await stream.send_message(response)

    async def __rpc_get_job_status_updates(
        self,
        stream: "grpclib.server.Stream[GetJobExecutionStatusRequest, GetJobExecutionStatusResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.get_job_status_updates,
            stream,
            request,
        )

    async def __rpc_paused_status(
        self, stream: "grpclib.server.Stream[PausedStatusRequest, PausedStatusResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.paused_status(request)
        await stream.send_message(response)

    async def __rpc_resume(
        self, stream: "grpclib.server.Stream[ResumeRequest, ResumeResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.resume(request)
        await stream.send_message(response)

    async def __rpc_halt(
        self, stream: "grpclib.server.Stream[HaltRequest, HaltResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.halt(request)
        await stream.send_message(response)

    async def __rpc_reset_data_processing(
        self,
        stream: "grpclib.server.Stream[ResetDataProcessingRequest, ResetDataProcessingResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.reset_data_processing(request)
        await stream.send_message(response)

    async def __rpc_validate_config(
        self,
        stream: "grpclib.server.Stream[_qua_config__.QuaConfig, _compiler__.ValidationResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.validate_config(request)
        await stream.send_message(response)

    async def __rpc_init(
        self,
        stream: "grpclib.server.Stream[betterproto_lib_google_protobuf.Empty, betterproto_lib_google_protobuf.Empty]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.init(request)
        await stream.send_message(response)

    async def __rpc_perform_qm_request(
        self,
        stream: "grpclib.server.Stream[_qm_api__.HighQmApiRequest, _qm_api__.HighQmApiResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.perform_qm_request(request)
        await stream.send_message(response)

    async def __rpc_request_data(
        self, stream: "grpclib.server.Stream[QmDataRequest, QmDataResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.request_data(request)
        await stream.send_message(response)

    async def __rpc_pull_simulator_samples(
        self,
        stream: "grpclib.server.Stream[_results_analyser__.PullSimulatorSamplesRequest, _results_analyser__.SimulatorSamplesResponse]",
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.pull_simulator_samples,
            stream,
            request,
        )

    async def __rpc_open_quantum_machine(
        self,
        stream: "grpclib.server.Stream[_qm_manager__.OpenQuantumMachineRequest, _qm_manager__.OpenQuantumMachineResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.open_quantum_machine(request)
        await stream.send_message(response)

    async def __rpc_close_quantum_machine(
        self,
        stream: "grpclib.server.Stream[_qm_manager__.CloseQuantumMachineRequest, _qm_manager__.CloseQuantumMachineResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.close_quantum_machine(request)
        await stream.send_message(response)

    async def __rpc_get_quantum_machine(
        self,
        stream: "grpclib.server.Stream[_qm_manager__.GetQuantumMachineRequest, _qm_manager__.GetQuantumMachineResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_quantum_machine(request)
        await stream.send_message(response)

    async def __rpc_get_running_job(
        self,
        stream: "grpclib.server.Stream[_qm_manager__.GetRunningJobRequest, _qm_manager__.GetRunningJobResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_running_job(request)
        await stream.send_message(response)

    async def __rpc_list_open_quantum_machines(
        self,
        stream: "grpclib.server.Stream[betterproto_lib_google_protobuf.Empty, _qm_manager__.ListOpenQuantumMachinesResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.list_open_quantum_machines(request)
        await stream.send_message(response)

    async def __rpc_close_all_quantum_machines(
        self,
        stream: "grpclib.server.Stream[betterproto_lib_google_protobuf.Empty, _qm_manager__.CloseAllQuantumMachinesResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.close_all_quantum_machines(request)
        await stream.send_message(response)

    async def __rpc_get_controllers(
        self,
        stream: "grpclib.server.Stream[betterproto_lib_google_protobuf.Empty, _qm_manager__.GetControllersResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_controllers(request)
        await stream.send_message(response)

    async def __rpc_is_job_running(
        self, stream: "grpclib.server.Stream[IsJobRunningRequest, IsJobRunningResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.is_job_running(request)
        await stream.send_message(response)

    async def __rpc_is_job_acquiring_data(
        self,
        stream: "grpclib.server.Stream[IsJobAcquiringDataRequest, IsJobAcquiringDataResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.is_job_acquiring_data(request)
        await stream.send_message(response)

    async def __rpc_simulate(
        self, stream: "grpclib.server.Stream[SimulationRequest, SimulationResponse]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.simulate(request)
        await stream.send_message(response)

    async def __rpc_clear_all_job_results(
        self,
        stream: "grpclib.server.Stream[betterproto_lib_google_protobuf.Empty, betterproto_lib_google_protobuf.Empty]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.clear_all_job_results(request)
        await stream.send_message(response)

    async def __rpc_perform_hal_debug_command(
        self,
        stream: "grpclib.server.Stream[PerformHalDebugCommandRequest, PerformHalDebugCommandResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.perform_hal_debug_command(request)
        await stream.send_message(response)

    async def __rpc_get_simulated_quantum_state(
        self,
        stream: "grpclib.server.Stream[GetSimulatedQuantumStateRequest, GetSimulatedQuantumStateResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_simulated_quantum_state(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/qm.grpc.frontend.Frontend/GetVersion": grpclib.const.Handler(
                self.__rpc_get_version,
                grpclib.const.Cardinality.UNARY_UNARY,
                betterproto_lib_google_protobuf.Empty,
                betterproto_lib_google_protobuf.StringValue,
            ),
            "/qm.grpc.frontend.Frontend/HealthCheck": grpclib.const.Handler(
                self.__rpc_health_check,
                grpclib.const.Cardinality.UNARY_UNARY,
                betterproto_lib_google_protobuf.Empty,
                HealthCheckResponse,
            ),
            "/qm.grpc.frontend.Frontend/Execute": grpclib.const.Handler(
                self.__rpc_execute,
                grpclib.const.Cardinality.UNARY_UNARY,
                ExecutionRequest,
                ExecutionResponse,
            ),
            "/qm.grpc.frontend.Frontend/AddToQueue": grpclib.const.Handler(
                self.__rpc_add_to_queue,
                grpclib.const.Cardinality.UNARY_UNARY,
                AddToQueueRequest,
                AddToQueueResponse,
            ),
            "/qm.grpc.frontend.Frontend/Compile": grpclib.const.Handler(
                self.__rpc_compile,
                grpclib.const.Cardinality.UNARY_UNARY,
                CompileRequest,
                CompileResponse,
            ),
            "/qm.grpc.frontend.Frontend/AddCompiledToQueue": grpclib.const.Handler(
                self.__rpc_add_compiled_to_queue,
                grpclib.const.Cardinality.UNARY_UNARY,
                AddCompiledToQueueRequest,
                AddCompiledToQueueResponse,
            ),
            "/qm.grpc.frontend.Frontend/RemovePendingJobs": grpclib.const.Handler(
                self.__rpc_remove_pending_jobs,
                grpclib.const.Cardinality.UNARY_UNARY,
                JobQueryParams,
                RemovePendingJobsResponse,
            ),
            "/qm.grpc.frontend.Frontend/GetPendingJobs": grpclib.const.Handler(
                self.__rpc_get_pending_jobs,
                grpclib.const.Cardinality.UNARY_UNARY,
                JobQueryParams,
                GetPendingJobsResponse,
            ),
            "/qm.grpc.frontend.Frontend/GetJobExecutionStatus": grpclib.const.Handler(
                self.__rpc_get_job_execution_status,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetJobExecutionStatusRequest,
                GetJobExecutionStatusResponse,
            ),
            "/qm.grpc.frontend.Frontend/GetJobStatusUpdates": grpclib.const.Handler(
                self.__rpc_get_job_status_updates,
                grpclib.const.Cardinality.UNARY_STREAM,
                GetJobExecutionStatusRequest,
                GetJobExecutionStatusResponse,
            ),
            "/qm.grpc.frontend.Frontend/PausedStatus": grpclib.const.Handler(
                self.__rpc_paused_status,
                grpclib.const.Cardinality.UNARY_UNARY,
                PausedStatusRequest,
                PausedStatusResponse,
            ),
            "/qm.grpc.frontend.Frontend/Resume": grpclib.const.Handler(
                self.__rpc_resume,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResumeRequest,
                ResumeResponse,
            ),
            "/qm.grpc.frontend.Frontend/Halt": grpclib.const.Handler(
                self.__rpc_halt,
                grpclib.const.Cardinality.UNARY_UNARY,
                HaltRequest,
                HaltResponse,
            ),
            "/qm.grpc.frontend.Frontend/ResetDataProcessing": grpclib.const.Handler(
                self.__rpc_reset_data_processing,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResetDataProcessingRequest,
                ResetDataProcessingResponse,
            ),
            "/qm.grpc.frontend.Frontend/ValidateConfig": grpclib.const.Handler(
                self.__rpc_validate_config,
                grpclib.const.Cardinality.UNARY_UNARY,
                _qua_config__.QuaConfig,
                _compiler__.ValidationResponse,
            ),
            "/qm.grpc.frontend.Frontend/Init": grpclib.const.Handler(
                self.__rpc_init,
                grpclib.const.Cardinality.UNARY_UNARY,
                betterproto_lib_google_protobuf.Empty,
                betterproto_lib_google_protobuf.Empty,
            ),
            "/qm.grpc.frontend.Frontend/PerformQmRequest": grpclib.const.Handler(
                self.__rpc_perform_qm_request,
                grpclib.const.Cardinality.UNARY_UNARY,
                _qm_api__.HighQmApiRequest,
                _qm_api__.HighQmApiResponse,
            ),
            "/qm.grpc.frontend.Frontend/RequestData": grpclib.const.Handler(
                self.__rpc_request_data,
                grpclib.const.Cardinality.UNARY_UNARY,
                QmDataRequest,
                QmDataResponse,
            ),
            "/qm.grpc.frontend.Frontend/PullSimulatorSamples": grpclib.const.Handler(
                self.__rpc_pull_simulator_samples,
                grpclib.const.Cardinality.UNARY_STREAM,
                _results_analyser__.PullSimulatorSamplesRequest,
                _results_analyser__.SimulatorSamplesResponse,
            ),
            "/qm.grpc.frontend.Frontend/OpenQuantumMachine": grpclib.const.Handler(
                self.__rpc_open_quantum_machine,
                grpclib.const.Cardinality.UNARY_UNARY,
                _qm_manager__.OpenQuantumMachineRequest,
                _qm_manager__.OpenQuantumMachineResponse,
            ),
            "/qm.grpc.frontend.Frontend/CloseQuantumMachine": grpclib.const.Handler(
                self.__rpc_close_quantum_machine,
                grpclib.const.Cardinality.UNARY_UNARY,
                _qm_manager__.CloseQuantumMachineRequest,
                _qm_manager__.CloseQuantumMachineResponse,
            ),
            "/qm.grpc.frontend.Frontend/GetQuantumMachine": grpclib.const.Handler(
                self.__rpc_get_quantum_machine,
                grpclib.const.Cardinality.UNARY_UNARY,
                _qm_manager__.GetQuantumMachineRequest,
                _qm_manager__.GetQuantumMachineResponse,
            ),
            "/qm.grpc.frontend.Frontend/GetRunningJob": grpclib.const.Handler(
                self.__rpc_get_running_job,
                grpclib.const.Cardinality.UNARY_UNARY,
                _qm_manager__.GetRunningJobRequest,
                _qm_manager__.GetRunningJobResponse,
            ),
            "/qm.grpc.frontend.Frontend/ListOpenQuantumMachines": grpclib.const.Handler(
                self.__rpc_list_open_quantum_machines,
                grpclib.const.Cardinality.UNARY_UNARY,
                betterproto_lib_google_protobuf.Empty,
                _qm_manager__.ListOpenQuantumMachinesResponse,
            ),
            "/qm.grpc.frontend.Frontend/CloseAllQuantumMachines": grpclib.const.Handler(
                self.__rpc_close_all_quantum_machines,
                grpclib.const.Cardinality.UNARY_UNARY,
                betterproto_lib_google_protobuf.Empty,
                _qm_manager__.CloseAllQuantumMachinesResponse,
            ),
            "/qm.grpc.frontend.Frontend/GetControllers": grpclib.const.Handler(
                self.__rpc_get_controllers,
                grpclib.const.Cardinality.UNARY_UNARY,
                betterproto_lib_google_protobuf.Empty,
                _qm_manager__.GetControllersResponse,
            ),
            "/qm.grpc.frontend.Frontend/IsJobRunning": grpclib.const.Handler(
                self.__rpc_is_job_running,
                grpclib.const.Cardinality.UNARY_UNARY,
                IsJobRunningRequest,
                IsJobRunningResponse,
            ),
            "/qm.grpc.frontend.Frontend/IsJobAcquiringData": grpclib.const.Handler(
                self.__rpc_is_job_acquiring_data,
                grpclib.const.Cardinality.UNARY_UNARY,
                IsJobAcquiringDataRequest,
                IsJobAcquiringDataResponse,
            ),
            "/qm.grpc.frontend.Frontend/Simulate": grpclib.const.Handler(
                self.__rpc_simulate,
                grpclib.const.Cardinality.UNARY_UNARY,
                SimulationRequest,
                SimulationResponse,
            ),
            "/qm.grpc.frontend.Frontend/ClearAllJobResults": grpclib.const.Handler(
                self.__rpc_clear_all_job_results,
                grpclib.const.Cardinality.UNARY_UNARY,
                betterproto_lib_google_protobuf.Empty,
                betterproto_lib_google_protobuf.Empty,
            ),
            "/qm.grpc.frontend.Frontend/PerformHalDebugCommand": grpclib.const.Handler(
                self.__rpc_perform_hal_debug_command,
                grpclib.const.Cardinality.UNARY_UNARY,
                PerformHalDebugCommandRequest,
                PerformHalDebugCommandResponse,
            ),
            "/qm.grpc.frontend.Frontend/GetSimulatedQuantumState": grpclib.const.Handler(
                self.__rpc_get_simulated_quantum_state,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetSimulatedQuantumStateRequest,
                GetSimulatedQuantumStateResponse,
            ),
        }



from qm._loc import _get_loc


class AnalogTimeDivision(object):
    def __init__(self, loc: str):
        self.loc = loc


class SlicedAnalogTimeDivision(AnalogTimeDivision):
    def __init__(self, loc: str, samples_per_chunk: int):
        super(SlicedAnalogTimeDivision, self).__init__(loc)
        self.samples_per_chunk = samples_per_chunk


class AccumulatedAnalogTimeDivision(AnalogTimeDivision):
    def __init__(self, loc: str, samples_per_chunk: int):
        super(AccumulatedAnalogTimeDivision, self).__init__(loc)
        self.samples_per_chunk = samples_per_chunk


class MovingWindowAnalogTimeDivision(AnalogTimeDivision):
    def __init__(self, loc: str, samples_per_chunk: int, chunks_per_window: int):
        super(MovingWindowAnalogTimeDivision, self).__init__(loc)
        self.samples_per_chunk = samples_per_chunk
        self.chunks_per_window = chunks_per_window


class AnalogProcessTarget(object):
    def __init__(self, loc: str):
        self.loc = loc


class ScalarProcessTarget(AnalogProcessTarget):
    def __init__(self, loc: str, target) -> None:
        super().__init__(loc)
        self.target = target


class VectorProcessTarget(AnalogProcessTarget):
    def __init__(self, loc: str, target, time_division: AnalogTimeDivision):
        AnalogProcessTarget.__init__(self, loc)
        self.target = target
        self.time_division = time_division


class AnalogMeasureProcess(object):
    def __init__(self, loc: str):
        self.loc = _get_loc()


class BareIntegration(AnalogMeasureProcess):
    def __init__(self, loc: str, element_output: str, iw: str, target: AnalogProcessTarget):
        super(BareIntegration, self).__init__(loc)
        self.element_output = element_output
        self.iw = iw
        self.target = target


class DualBareIntegration(AnalogMeasureProcess):
    def __init__(
        self,
        loc: str,
        element_output1: str,
        element_output2: str,
        iw1: str,
        iw2: str,
        target: AnalogProcessTarget,
    ):
        super(DualBareIntegration, self).__init__(loc)
        self.element_output1 = element_output1
        self.element_output2 = element_output2
        self.iw1 = iw1
        self.iw2 = iw2
        self.target = target


class DemodIntegration(AnalogMeasureProcess):
    def __init__(self, loc: str, element_output: str, iw: str, target: AnalogProcessTarget):
        super(DemodIntegration, self).__init__(loc)
        self.element_output = element_output
        self.iw = iw
        self.target = target


class DualDemodIntegration(AnalogMeasureProcess):
    def __init__(
        self,
        loc: str,
        element_output1: str,
        element_output2: str,
        iw1: str,
        iw2: str,
        target: AnalogProcessTarget,
    ):
        super(DualDemodIntegration, self).__init__(loc)
        self.element_output1 = element_output1
        self.element_output2 = element_output2
        self.iw1 = iw1
        self.iw2 = iw2
        self.target = target


class RawTimeTagging(AnalogMeasureProcess):
    def __init__(self, loc: str, element_output: str, target, targetLen, max_time):
        super(RawTimeTagging, self).__init__(loc)
        self.element_output = element_output
        self.target = target
        self.targetLen = targetLen
        self.max_time = max_time


class HighResTimeTagging(AnalogMeasureProcess):
    def __init__(self, loc: str, element_output: str, target, targetLen, max_time):
        super(HighResTimeTagging, self).__init__(loc)
        self.element_output = element_output
        self.target = target
        self.targetLen = targetLen
        self.max_time = max_time



import logging
import warnings
import contextlib
import dataclasses
from enum import Enum
from time import perf_counter
from abc import abstractmethod
from functools import lru_cache
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Set, Dict, List, Tuple, Union, Generic, TypeVar, Iterator, Optional, cast

import betterproto
from octave_sdk.octave import ClockInfo
from octave_sdk._octave_client import MonitorResult
from octave_sdk.grpc.quantummachines.octave.api.v1 import (
    ClockUpdate,
    OctaveModule,
    SingleUpdate,
    RfUpConvUpdate,
    ClockUpdateMode,
)
from octave_sdk import (
    IFMode,
    Octave,
    ClockType,
    OctaveOutput,
    RFOutputMode,
    ClockFrequency,
    OctaveLOSource,
    RFInputLOSource,
    RFInputRFSource,
)

from qm.type_hinting import Number
from qm.utils.config_utils import get_fem_config
from qm.octave.octave_config import QmOctaveConfig
from qm.type_hinting.config_types import StandardPort
from qm.octave._calibration_config import _prep_config
from qm.program._qua_config_to_pb import build_iw_sample
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.server_details import ConnectionDetails
from qm.exceptions import OpenQmException, OctaveLoopbackError, NoOutputPortDeclared
from qm.octave._calibration_names import COMMON_OCTAVE_PREFIX, CalibrationElementsNames
from qm.octave.octave_mixer_calibration import (
    AutoCalibrationParams,
    DeprecatedCalibrationResult,
    convert_to_old_calibration_result,
)
from qm.grpc.qua_config import (
    QuaConfig,
    QuaConfigMatrix,
    QuaConfigMixerDec,
    QuaConfigPulseDec,
    QuaConfigMixInputs,
    QuaConfigElementDec,
    QuaConfigSingleInput,
    QuaConfigWaveformDec,
    QuaConfigOctaveConfig,
    QuaConfigControllerDec,
    QuaConfigOctoDacFemDec,
    QuaConfigOctaveLoopback,
    QuaConfigCorrectionEntry,
    QuaConfigAdcPortReference,
    QuaConfigDacPortReference,
    QuaConfigPulseDecOperation,
    QuaConfigDigitalWaveformDec,
    QuaConfigConstantWaveformDec,
    QuaConfigOctaveLoSourceInput,
    QuaConfigIntegrationWeightDec,
    QuaConfigDigitalWaveformSample,
    QuaConfigOctaveOutputSwitchState,
    QuaConfigOctaveDownconverterRfSource,
)

if TYPE_CHECKING:
    from qm.api.v2.qm_api import QmApi
    from qm.quantum_machine import QuantumMachine
    from qm.quantum_machines_manager import QuantumMachinesManager

logger = logging.getLogger(__name__)


class SetFrequencyException(Exception):
    pass


class MixerCalibrationException(Exception):
    pass


class ClockMode(Enum):
    Internal = (ClockType.Internal, None)
    External_10MHz = (ClockType.External, ClockFrequency.MHZ_10)
    External_100MHz = (ClockType.External, ClockFrequency.MHZ_100)
    External_1000MHz = (ClockType.Buffered, ClockFrequency.MHZ_1000)


_PortType = TypeVar("_PortType", QuaConfigDacPortReference, QuaConfigAdcPortReference)


@dataclasses.dataclass
class ControllerPorts(Generic[_PortType]):
    I: _PortType
    Q: _PortType

    def valid(self, pb_config: QuaConfig) -> bool:
        if self.I is None or self.Q is None:
            return False
        if self.I.controller != self.Q.controller:
            return False
        if self.I.fem != self.Q.fem:
            return False
        try:
            controller_config = get_fem_config(pb_config, self.I)
        except KeyError:
            return False
        relevant_ports = self._get_relevant_ports(controller_config)
        return self.I.number in relevant_ports and self.Q.number in relevant_ports

    @abstractmethod
    def _get_relevant_ports(self, controller_config: Union[QuaConfigControllerDec, QuaConfigOctoDacFemDec]) -> Set[int]:
        pass


class _ControllerDACPorts(ControllerPorts[QuaConfigDacPortReference]):
    def _get_relevant_ports(self, controller_config: Union[QuaConfigControllerDec, QuaConfigOctoDacFemDec]) -> Set[int]:
        return set(controller_config.analog_outputs)


class _ControllerADCPorts(ControllerPorts[QuaConfigAdcPortReference]):
    def _get_relevant_ports(self, controller_config: Union[QuaConfigControllerDec, QuaConfigOctoDacFemDec]) -> Set[int]:
        return set(controller_config.analog_inputs)


@dataclasses.dataclass
class OctaveConnection:
    dacs: _ControllerDACPorts
    adcs: _ControllerADCPorts

    def valid(self, pb_config: QuaConfig) -> bool:
        return self.dacs.valid(pb_config) and self.adcs.valid(pb_config)


@dataclasses.dataclass
class _UpconvertedState:
    lo_source: Optional[OctaveLOSource] = None
    lo_frequency: Optional[float] = None
    output_mode: Optional[RFOutputMode] = None
    output_gain: Optional[float] = None


class OctaveManager:
    def __init__(
        self,
        config: Optional[QmOctaveConfig],
        qmm: "QuantumMachinesManager",
        capabilities: ServerCapabilities,
    ) -> None:

        self._qmm = qmm
        self._capabilities = capabilities
        self._octave_config = config or QmOctaveConfig()
        self._upconverted_states: Dict[Tuple[str, int], _UpconvertedState] = {}
        self._perform_healthcheck_at_init()

    def _perform_healthcheck_at_init(self) -> None:
        for octave_name in self._octave_config.get_devices():
            self.get_client(octave_name)

    def get_client(self, name: str) -> Octave:
        return get_device(
            connection_info=self._octave_config._devices[name],
            loop_backs=self._octave_config.get_lo_loopbacks_by_octave(name),
            octave_name=name,
            fan=self._octave_config.fan,
        )

    def get_output_port(
        self,
        opx_i_port: StandardPort,
        opx_q_port: StandardPort,
    ) -> Optional[Tuple[str, int]]:
        warnings.warn(
            "This function is deprecated since 1.1.6 and will be removed in 1.2.0.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._octave_config.get_octave_input_port(opx_i_port, opx_q_port)

    def _get_client_from_port(self, octave_port: Tuple[str, int]) -> Octave:
        return self.get_client(octave_port[0])

    def restore_default_state(self, octave_name: str) -> None:
        self.get_client(octave_name).restore_default_state()

    def start_batch_mode(self) -> None:
        for octave_name in self._octave_config.get_devices():
            self.get_client(octave_name).start_batch_mode()

    def end_batch_mode(self) -> None:
        for octave_name in self._octave_config.get_devices():
            self.get_client(octave_name).end_batch_mode()

    @contextlib.contextmanager
    def batch_mode(self) -> Iterator[None]:
        self.start_batch_mode()
        yield
        self.end_batch_mode()

    def set_clock(
        self,
        octave_name: str,
        clock_type: Optional[ClockType] = None,
        frequency: Optional[ClockFrequency] = None,
        clock_mode: Optional[ClockMode] = None,
    ) -> None:

        """This function will set the octave clock type - internal, external or buffered.
        It can also set the clock frequency - 10, 100 or 1000 MHz

        Args:
            octave_name: str
            clock_type: ClockType
            frequency: ClockFrequency
            clock_mode: ClockMode

        Returns:

        """
        if clock_mode is None:
            pass
            # warnings.warn(
            #     "set_clock is changing its API, and the 'clock_type' and 'frequency' arguments will be "
            #     "removed in the next version, please use the 'mode' parameter.",
            #     category=DeprecationWarning,
            #     stacklevel=2,
            # )
        else:
            clock_type, frequency = clock_mode.value

        client = self.get_client(octave_name)
        client.set_clock(clock_type, frequency)
        client.save_default_state(only_clock=True)

    def get_clock(self, octave_name: str) -> ClockInfo:
        """Return the octave clock type and frequency

        Args:
            octave_name: str

        Returns:
            ClockInfo
        """
        clock = self.get_client(octave_name).get_clock()
        return ClockInfo(clock.clock_type, clock.frequency)

    def set_lo_frequency(
        self,
        octave_output_port: Tuple[str, int],
        lo_frequency: float,
        lo_source: OctaveLOSource = OctaveLOSource.Internal,
        set_source: bool = True,
    ) -> None:
        """Sets the LO frequency of the synthesizer associated to element

        Args:
            octave_output_port
            lo_frequency
            lo_source
            set_source
        """
        # warnings.warn(
        #     deprecation_message(
        #         "OctaveManager.set_lo_frequency()",
        #         "1.1.0",
        #         "1.2.0",
        #         "set_lo_frequency will be removed, please use the relevant element.",
        #     ),
        #     category=DeprecationWarning,
        # )
        octave = self._get_client_from_port(octave_output_port)
        octave_name, port_index = octave_output_port

        loop_backs = self._octave_config.get_lo_loopbacks_by_octave(octave_name)

        if lo_source != OctaveLOSource.Internal and lo_source not in loop_backs:
            raise SetFrequencyException(f"Cannot set frequency to an external lo source" f" {lo_source.name}")

        if set_source:
            octave.rf_outputs[port_index].set_lo_source(lo_source)

        octave.rf_outputs[port_index].set_lo_frequency(lo_source, lo_frequency)

    def set_lo_source(self, octave_output_port: Tuple[str, int], lo_port: OctaveLOSource) -> None:
        """Sets the source of LO going to the upconverter associated with element.

        Args:
            octave_output_port
            lo_port
        """
        octave = self._get_client_from_port(octave_output_port)
        octave.rf_outputs[octave_output_port[1]].set_lo_source(lo_port)

    def set_rf_output_mode(self, octave_output_port: Tuple[str, int], switch_mode: RFOutputMode) -> None:
        """Configures the output switch of the upconverter associated to element.
        switch_mode can be either: 'always_on', 'always_off', 'normal' or 'inverted'
        When in 'normal' mode a high trigger will turn the switch on and a low
        trigger will turn it off
        When in 'inverted' mode a high trigger will turn the switch off and a low
        trigger will turn it on
        When in 'always_on' the switch will be permanently on. When in 'always_off'
        mode the switch will be permanently off.

        Args:
            octave_output_port
            switch_mode
        """
        octave = self._get_client_from_port(octave_output_port)
        index = octave_output_port[1]
        octave.rf_outputs[octave_output_port[1]].set_output(switch_mode)

        # Shuts down the second stage amplifier if the switch is off
        power_amp_enabled = switch_mode != RFOutputMode.off
        octave._client.update(
            updates=[SingleUpdate(rf_up_conv=RfUpConvUpdate(index=index, power_amp_enabled=power_amp_enabled))]
        )

    def set_rf_output_gain(
        self,
        octave_output_port: Tuple[str, int],
        gain_in_db: float,
        lo_frequency: Optional[float] = None,
    ) -> None:
        """Sets the RF output gain for the upconverter associated with element.
        if no lo_frequency is given, and lo source is internal, will use the
        internal frequency

        Args:
            octave_output_port
            gain_in_db
            lo_frequency
        """
        octave = self._get_client_from_port(octave_output_port)
        octave.rf_outputs[octave_output_port[1]].set_gain(gain_in_db, lo_frequency)

    def set_downconversion_lo_source(
        self,
        octave_input_port: Tuple[str, int],
        lo_source: RFInputLOSource,
        lo_frequency: Optional[float] = None,
        disable_warning: Optional[bool] = False,
    ) -> None:
        """Sets the LO source for the downconverters.

        Args:
            octave_input_port
            lo_source
            lo_frequency
            disable_warning
        """
        octave = self._get_client_from_port(octave_input_port)
        octave.rf_inputs[octave_input_port[1]].set_lo_source(lo_source)
        octave.rf_inputs[octave_input_port[1]].set_rf_source(RFInputRFSource.RF_in)
        internal = lo_source == RFInputLOSource.Internal or lo_source == RFInputLOSource.Analyzer
        if lo_frequency is not None and internal:
            octave.rf_inputs[octave_input_port[1]].set_lo_frequency(source_name=lo_source, frequency=lo_frequency)

    def set_downconversion_if_mode(
        self,
        octave_input_port: Tuple[str, int],
        if_mode_i: IFMode = IFMode.direct,
        if_mode_q: IFMode = IFMode.direct,
        disable_warning: Optional[bool] = False,
    ) -> None:
        """Sets the IF downconversion stage.
        if_mode can be one of: 'direct', 'mixer', 'envelope_DC', 'envelope_AC','OFF'
        If only one value is given the setting is applied to both IF channels
        (I and Q) for the downconverter associated to element
        (how will we know that? shouldn't this be per downconverter?)
        If if_mode is a tuple, then the IF stage will be assigned to each
        quadrature independently, i.e.:
        if_mode = ('direct', 'envelope_AC') will set the I-channel to be
        direct and the Q-channel to be 'envelope_AC'

        Args:
            disable_warning
            octave_input_port
            if_mode_q
            if_mode_i
        """

        octave = self._get_client_from_port(octave_input_port)
        octave.rf_inputs[octave_input_port[1]].set_if_mode_i(if_mode_i)
        octave.rf_inputs[octave_input_port[1]].set_if_mode_q(if_mode_q)

    def reset(self, octave_name: str) -> bool:
        """
        Will reset the entire Octave HW to default off state
        Warning, will block the code until reset completes

        Args:
            octave_name: str
        Returns:
            True on success, False otherwise
        """
        if self._capabilities.supports_octave_reset:
            return cast(bool, self.get_client(octave_name).reset())
        else:
            logger.error("QOP version do not support Octave reset function")
            return False

    def calibrate(
        self,
        octave_output_port: Tuple[str, int],
        lo_if_frequencies_tuple_list: List[Tuple[int, int]],
        save_to_db: bool = True,
        close_open_quantum_machines: bool = True,
        optimizer_parameters: Optional[Dict[str, Any]] = None,
        offset_frequency: float = 7e6,
        **kwargs: Any,
    ) -> Dict[Tuple[int, int], DeprecatedCalibrationResult]:
        """calibrates IQ mixer associated with element

        Args:
            close_open_quantum_machines: Boolean, if true (default) all
                running QMs
            octave_output_port
            lo_if_frequencies_tuple_list: A list of LO/IF frequencies
                for which the calibration is to be performed [(LO1,
                IF1), (LO2, IF2), ...]
            save_to_db
            optimizer_parameters
        will be closed for the calibration. Otherwise,
        calibration might fail if there are not enough resources for the calibration
        """
        # warnings.warn(
        #     "This function is deprecated, please use the 'calibrate_element' in the QuantumMachine instance",
        #     category=DeprecationWarning,
        # )
        if kwargs:
            logger.warning(f"unused kwargs: {list(kwargs)}, please remove them.")

        lo_to_if_mapping = _create_lo_to_if_mapping(lo_if_frequencies_tuple_list)
        lo_freq = list(lo_to_if_mapping)[0]

        if_freq_list = lo_to_if_mapping[lo_freq]
        first_if = if_freq_list[0]

        qm_inst = self._create_dedicated_qm(
            lo_freq,
            first_if,
            octave_output_port,
            close_open_quantum_machines,
        )

        calibration_result = qm_inst.calibrate_element(
            "to_calibrate",
            lo_to_if_mapping,
            save_to_db=save_to_db,
            params=AutoCalibrationParams(offset_frequency=offset_frequency),
        )

        return convert_to_old_calibration_result(calibration_result, "Correction_mixer")

    def _create_dedicated_qm(
        self,
        first_lo: Number,
        first_if: Number,
        octave_output_port: Tuple[str, int],
        close_open_quantum_machines: bool = True,
    ) -> Union["QuantumMachine", "QmApi"]:
        iq_channels = self._octave_config.get_opx_iq_ports(octave_output_port)
        controller_name = iq_channels[0][0]
        adc_channels = ((controller_name, 1), (controller_name, 2))
        config = _prep_config(
            iq_channels,
            adc_channels,
            first_if,
            first_lo,
        )
        try:
            t0 = perf_counter()
            qm_inst = self._qmm.open_qm(
                config, close_other_machines=close_open_quantum_machines, add_calibration_elements_to_config=True
            )
            logger.debug(f"Creating dedicated QM for calibration took {perf_counter() - t0}")
        except OpenQmException as e:
            raise MixerCalibrationException("Mixer calibration failed: Could not open a quantum machine.") from e

        return qm_inst

    def set_octaves_from_qua_config(self, octaves_config: Dict[str, QuaConfigOctaveConfig]) -> None:
        if not octaves_config:
            # This if serves two things - if the user uses the old API, no need to set the octaves,
            # and if there are no octaves in the qua config, but there are in the cluster,
            # we still want to skip the batch mode.
            return
        for octave_name, octave_config in octaves_config.items():
            pb_loopbacks = octave_config.loopbacks
            loopbacks = get_loopbacks_from_pb(pb_loopbacks, octave_name)
            connection_info = self._octave_config.devices[octave_name]
            octave = get_device(connection_info, loopbacks, octave_name)
            octave.start_batch_mode()
            for output_port_idx, output_config in octave_config.rf_outputs.items():
                output_client = octave.rf_outputs[output_port_idx]
                config_lo_source = output_config.lo_source
                if config_lo_source == QuaConfigOctaveLoSourceInput.internal:
                    lo_source = OctaveLOSource.Internal
                elif config_lo_source == QuaConfigOctaveLoSourceInput.external:
                    lo_source = OctaveLOSource[f"LO{output_port_idx}"]
                else:
                    raise ValueError(f"lo_source {config_lo_source} is not supported")

                output_client.set_lo_source(lo_source)

                if lo_source == OctaveLOSource.Internal or lo_source in loopbacks:
                    output_client.set_lo_frequency(lo_source, output_config.lo_frequency)
                else:
                    logger.debug(f"Cannot set frequency to an external lo source {lo_source.name}")

                output_mode = {
                    QuaConfigOctaveOutputSwitchState.always_on: RFOutputMode.on,
                    QuaConfigOctaveOutputSwitchState.always_off: RFOutputMode.off,
                    QuaConfigOctaveOutputSwitchState.triggered: RFOutputMode.trig_normal,
                    QuaConfigOctaveOutputSwitchState.triggered_reversed: RFOutputMode.trig_inverse,
                }[output_config.output_mode]
                output_client.set_output(output_mode)
                output_client.set_gain(
                    output_config.gain,
                    output_config.lo_frequency,
                    use_iq_attenuators=output_config.input_attenuators,
                )

            for input_port_idx, input_config in octave_config.rf_inputs.items():
                input_client = octave.rf_inputs[input_port_idx]
                input_source = {
                    QuaConfigOctaveDownconverterRfSource.rf_in: RFInputRFSource.RF_in,
                    QuaConfigOctaveDownconverterRfSource.loopback_1: RFInputRFSource.Loopback_RF_out_1,
                    QuaConfigOctaveDownconverterRfSource.loopback_2: RFInputRFSource.Loopback_RF_out_2,
                    QuaConfigOctaveDownconverterRfSource.loopback_3: RFInputRFSource.Loopback_RF_out_3,
                    QuaConfigOctaveDownconverterRfSource.loopback_4: RFInputRFSource.Loopback_RF_out_4,
                    QuaConfigOctaveDownconverterRfSource.loopback_5: RFInputRFSource.Loopback_RF_out_5,
                }[input_config.rf_source]
                input_client.set_rf_source(input_source)

                if input_config.lo_source == QuaConfigOctaveLoSourceInput.internal:
                    downconversion_lo_source = RFInputLOSource.Internal
                elif input_config.lo_source == QuaConfigOctaveLoSourceInput.external:
                    downconversion_lo_source = RFInputLOSource[f"Dmd{input_port_idx}LO"]
                elif input_config.lo_source == QuaConfigOctaveLoSourceInput.analyzer:
                    downconversion_lo_source = RFInputLOSource.Analyzer
                else:
                    raise ValueError(f"lo_source {input_config.lo_source} is not supported")

                input_client.set_lo_source(downconversion_lo_source)
                if (
                    downconversion_lo_source == RFInputLOSource.Internal
                    or (downconversion_lo_source == RFInputLOSource.Dmd1LO and OctaveLOSource.Dmd1LO in loopbacks)
                    or (downconversion_lo_source == RFInputLOSource.Dmd2LO and OctaveLOSource.Dmd2LO in loopbacks)
                ):
                    input_client.set_lo_frequency(downconversion_lo_source, input_config.lo_frequency)
                else:
                    logger.debug(f"Cannot set frequency to an external lo source {downconversion_lo_source.name}")

                input_client.set_if_mode_i(IFMode[input_config.if_mode_i.name])
                input_client.set_if_mode_q(IFMode[input_config.if_mode_q.name])
            octave.end_batch_mode()


def get_som_temp(monitor_data: MonitorResult) -> float:
    return cast(float, monitor_data.modules[OctaveModule.SOM][0].temp)


def _set_clock_in(octave: Octave, clock_in: ClockInfo) -> None:
    client = octave._client

    def _update_clock(mode: ClockUpdateMode, frequency: float) -> None:
        _update = ClockUpdate(mode=mode, clock_frequency=frequency, synthesizers_clock=125e6)
        client.update(updates=[SingleUpdate(clock=_update)])

    _update_clock(ClockUpdateMode.MODE_INTERNAL_USE_1G, 10e6)

    mon = client.monitor()
    has_1g = 0 < get_som_temp(mon) < 100

    if clock_in.clock_type == ClockType.Internal:
        if has_1g:
            # we are already have a programmed internal clock (using the internal 1G)
            return

        # An internal clock is already programmed, but it seems like the internal 1G VCO is
        # not working. Hence, we switch to use the internal 3G VCO (that's the best we can do)
        _update_clock(ClockUpdateMode.MODE_INTERNAL, 10e6)
        return

    if clock_in.clock_type == ClockType.Buffered:
        _update_clock(ClockUpdateMode.MODE_BUFFERED, 1e9)

        # The FPGA's temperature is our indication that the clock is fine. We read it again after
        # changing the clock configuration.
        mon = client.monitor()
        if 0 < get_som_temp(mon) < 100:
            # All good, the buffered clock feeds the FPGA successfully.
            return

        clock_mode = ClockUpdateMode.MODE_INTERNAL_USE_1G
        if not has_1g:
            clock_mode = ClockUpdateMode.MODE_INTERNAL
        _update_clock(clock_mode, 10e6)
        print("Not sensing any 1GHz input clock for the octave (buffered), switching to internal clock")
        return

    # Clock mode is external at this point
    clock_mode = ClockUpdateMode.MODE_EXTERNAL_USE_1G
    if not has_1g:
        clock_mode = ClockUpdateMode.MODE_EXTERNAL
    clock_in_frequency = {
        ClockFrequency.MHZ_10: 10e6,
        ClockFrequency.MHZ_100: 100e6,
        ClockFrequency.MHZ_1000: 1000e6,
    }[clock_in.frequency]
    _update_clock(clock_mode, clock_in_frequency)

    # Again, verify the clock is good...
    mon = client.monitor()
    if 0 < get_som_temp(mon) < 100:
        # All good, the buffered clock feeds the FPGA successfully.
        return

    clock_mode = ClockUpdateMode.MODE_INTERNAL_USE_1G
    if not has_1g:
        clock_mode = ClockUpdateMode.MODE_INTERNAL
    _update_clock(clock_mode, 10e6)
    print("Not sensing any 1GHz input clock for the octave (external), switching to internal clock")


def create_lo_to_if_list_mapping(lo_if_frequencies_tuple_list: List[Tuple[int, int]]) -> Dict[int, List[int]]:
    lo_to_if_mapping = defaultdict(list)
    for lo_freq, if_freq in lo_if_frequencies_tuple_list:
        lo_to_if_mapping[lo_freq].append(if_freq)
    return lo_to_if_mapping


def get_loopbacks_from_pb(
    pb_loopbacks: List[QuaConfigOctaveLoopback], octave_name: str
) -> Dict[OctaveLOSource, OctaveOutput]:
    loopbacks = {}
    for loopback in pb_loopbacks:
        source_octave = loopback.lo_source_generator.device_name
        if source_octave != octave_name:
            raise OctaveLoopbackError
        lo_source_input = OctaveLOSource[loopback.lo_source_input.name]
        output_name = loopback.lo_source_generator.port_name.name
        standardized_name = output_name[0].upper() + output_name[1:]
        # we use snake case but octave_sdk is camel case
        loopbacks[lo_source_input] = OctaveOutput[standardized_name]
    return loopbacks


OctaveConnectionsType = Dict[str, Dict[int, OctaveConnection]]


def _add_octave_connections_from_octave_config(octave_config: QmOctaveConfig) -> OctaveConnectionsType:
    all_octave_connections: OctaveConnectionsType = defaultdict(dict)

    opx_output_octave_port_mapping = octave_config.get_opx_octave_port_mapping()
    octave_opx_output_port_mapping = {v: k for k, v in opx_output_octave_port_mapping.items()}
    octave_names = {v[0] for v in octave_opx_output_port_mapping}
    octave_opx_input_port_mapping = octave_config.get_octave_to_opx_port_mapping()
    for octave_name in octave_names:
        for channel_index in range(1, 6):
            i_key, q_key = (octave_name, f"I{channel_index}"), (octave_name, f"Q{channel_index}")
            if i_key in octave_opx_output_port_mapping and q_key in octave_opx_output_port_mapping:
                con_i, fem_i, idx_i = octave_opx_output_port_mapping[i_key]
                con_q, fem_q, idx_q = octave_opx_output_port_mapping[q_key]
                dac_ports = _ControllerDACPorts(
                    I=QuaConfigDacPortReference(controller=con_i, fem=fem_i, number=idx_i),
                    Q=QuaConfigDacPortReference(controller=con_q, fem=fem_q, number=idx_q),
                )

                i_key_in, q_key_in = (octave_name, "I"), (octave_name, "Q")
                con_i_in, fem_i_in, idx_i_in = octave_opx_input_port_mapping.get(i_key_in, (con_i, fem_i, 1))
                con_q_in, fem_q_in, idx_q_in = octave_opx_input_port_mapping.get(q_key_in, (con_q, fem_q, 2))
                adc_ports = _ControllerADCPorts(
                    I=QuaConfigAdcPortReference(controller=con_i_in, fem=fem_i_in, number=idx_i_in),
                    Q=QuaConfigAdcPortReference(controller=con_q_in, fem=fem_q_in, number=idx_q_in),
                )
                all_octave_connections[octave_name][channel_index] = OctaveConnection(dacs=dac_ports, adcs=adc_ports)
    return dict(all_octave_connections)


def _add_octave_connections_from_pb_config(pb_config: QuaConfig) -> OctaveConnectionsType:
    all_octave_connections: OctaveConnectionsType = defaultdict(dict)
    for octave_name, octave_pb_config in pb_config.v1_beta.octaves.items():
        if betterproto.serialized_on_wire(octave_pb_config.if_outputs):
            adc_ports = _ControllerADCPorts(
                I=octave_pb_config.if_outputs.if_out1.port,
                Q=octave_pb_config.if_outputs.if_out2.port,
            )
            for upconverter_index, upconverter_config in octave_pb_config.rf_outputs.items():
                dac_ports = _ControllerDACPorts(
                    I=upconverter_config.i_connection,
                    Q=upconverter_config.q_connection,
                )
                connections = OctaveConnection(dacs=dac_ports, adcs=adc_ports)
                all_octave_connections[octave_name][upconverter_index] = connections
    return dict(all_octave_connections)


def _filter_only_valid_connections(
    octave_connections: OctaveConnectionsType, pb_config: QuaConfig
) -> OctaveConnectionsType:
    filtered_connections: OctaveConnectionsType = defaultdict(dict)
    for octave_name, curr_connections in octave_connections.items():
        for upconverter_index, connections in curr_connections.items():
            if connections.valid(pb_config):
                filtered_connections[octave_name][upconverter_index] = connections
    return dict(filtered_connections)


def _get_octave_channels_to_opx_ports(
    pb_config: QuaConfig, octave_config: QmOctaveConfig
) -> Dict[str, Dict[int, OctaveConnection]]:
    all_octave_connections = _add_octave_connections_from_octave_config(octave_config)
    all_octave_connections.update(_add_octave_connections_from_pb_config(pb_config))
    return _filter_only_valid_connections(all_octave_connections, pb_config)


def prep_config_for_calibration(
    pb_config: QuaConfig, octave_config: QmOctaveConfig, capabilities: ServerCapabilities
) -> QuaConfig:
    all_octave_connections = _get_octave_channels_to_opx_ports(pb_config, octave_config)
    if not all_octave_connections:
        logger.warning(
            "No valid channels found for calibration, make sure you connect I and Q to the same device, "
            "and that you declare the relevant ports in the config."
        )
        return pb_config

    return _add_calibration_entries_to_config(pb_config, capabilities, all_octave_connections)


def _find_dummy_controller_outputs(
    adc: QuaConfigAdcPortReference, pb_config: QuaConfig, all_octave_connections: OctaveConnectionsType
) -> Tuple[QuaConfigDacPortReference, QuaConfigDacPortReference]:
    controller_config = get_fem_config(pb_config, adc)
    port_candidates = list(controller_config.analog_outputs)
    if len(port_candidates) < 2:
        raise NoOutputPortDeclared(
            f"Could not find a controller outputs for input port {adc}. "
            f"For now, at least two output ports of the measuring controller must be declared"
        )
    controller, fem = adc.controller, adc.fem
    # This loop comes so the chosen ports will not be arbitrary and raise "cable swap error".
    # So we take ports that are connected correctly to some octave
    for octave_connections in all_octave_connections.values():
        for channel_connections in octave_connections.values():
            cond_i = channel_connections.dacs.I.controller == controller and channel_connections.dacs.I.fem == fem
            cond_q = channel_connections.dacs.Q.controller == controller and channel_connections.dacs.Q.fem == fem
            ports_declared = {channel_connections.dacs.I.number, channel_connections.dacs.Q.number} <= set(
                port_candidates
            )
            if cond_i and cond_q and ports_declared:
                return channel_connections.dacs.I, channel_connections.dacs.Q

    # If we don't find such, we take arbitrary ports.
    return (
        QuaConfigDacPortReference(controller=controller, fem=fem, number=port_candidates[0]),
        QuaConfigDacPortReference(controller=controller, fem=fem, number=port_candidates[1]),
    )


def _add_calibration_entries_to_config(
    pb_config: QuaConfig, capabilities: ServerCapabilities, all_octave_connections: OctaveConnectionsType
) -> QuaConfig:

    dummy_lo_frequency = int(6e9)
    dummy_if_frequency: float = 50e6

    dummy_down_mixer_offset: float = 7e6

    offset_amp = 0.25
    weight_amplitude = 1.0

    integration_length = 10000
    stabilization_length = 1400
    wrap_up_length = 600

    calibration_pulse_length = integration_length + stabilization_length + wrap_up_length
    calibration_amp = 0.125

    time_of_flight = stabilization_length

    frequency_idx = int(capabilities.supports_double_frequency)  # This is to shorten many if-else in the code using
    # a trinary expression
    dummy_mixer_name = f"{COMMON_OCTAVE_PREFIX}dummy_mixer"
    dummy_lo_frequency_double = [0.0, float(dummy_lo_frequency)][frequency_idx]

    for octave_name, octave_connections in all_octave_connections.items():
        for upconverter_index, channel_connections in octave_connections.items():
            names = CalibrationElementsNames(octave_name, upconverter_index)

            def _create_mix_inputs(set_dummy_ports: bool) -> QuaConfigMixInputs:
                if set_dummy_ports:
                    i, q = _find_dummy_controller_outputs(channel_connections.adcs.I, pb_config, all_octave_connections)
                else:
                    i, q = channel_connections.dacs.I, channel_connections.dacs.Q

                return QuaConfigMixInputs(
                    i=i,
                    q=q,
                    lo_frequency=dummy_lo_frequency,
                    lo_frequency_double=dummy_lo_frequency_double,
                    mixer=dummy_mixer_name,
                )

            def create_analyzer_element(intermediate_frequency: float) -> QuaConfigElementDec:

                return QuaConfigElementDec(
                    intermediate_frequency=abs(int(intermediate_frequency)),
                    intermediate_frequency_double=[0.0, abs(float(intermediate_frequency))][frequency_idx],
                    intermediate_frequency_negative=intermediate_frequency < 0,
                    mix_inputs=_create_mix_inputs(True),
                    operations={"Analyze": f"{COMMON_OCTAVE_PREFIX}Analyze_pulse"},
                    outputs={"out1": channel_connections.adcs.I, "out2": channel_connections.adcs.Q},
                    time_of_flight=time_of_flight,
                    smearing=0,
                )

            i_offset = QuaConfigElementDec(
                single_input=QuaConfigSingleInput(port=channel_connections.dacs.I),
                operations={"DC_offset": f"{COMMON_OCTAVE_PREFIX}DC_offset_pulse"},
            )
            q_offset = QuaConfigElementDec(
                single_input=QuaConfigSingleInput(port=channel_connections.dacs.Q),
                operations={"DC_offset": f"{COMMON_OCTAVE_PREFIX}DC_offset_pulse"},
            )
            pb_config.v1_beta.elements.update(
                {
                    names.iq_mixer: QuaConfigElementDec(
                        mix_inputs=_create_mix_inputs(False),
                        intermediate_frequency=abs(int(dummy_if_frequency)),
                        intermediate_frequency_double=[0.0, abs(float(dummy_if_frequency))][frequency_idx],
                        operations={"calibration": f"{COMMON_OCTAVE_PREFIX}calibration_pulse"},
                        digital_inputs={},
                    ),
                    names.i_offset: i_offset,
                    names.q_offset: q_offset,
                    names.signal_analyzer: create_analyzer_element(dummy_if_frequency - dummy_down_mixer_offset),
                    names.lo_analyzer: create_analyzer_element(-dummy_down_mixer_offset),
                    names.image_analyzer: create_analyzer_element(-dummy_if_frequency - dummy_down_mixer_offset),
                }
            )

    pb_config.v1_beta.pulses.update(
        {
            f"{COMMON_OCTAVE_PREFIX}calibration_pulse": QuaConfigPulseDec(
                operation=QuaConfigPulseDecOperation.CONTROL,
                length=calibration_pulse_length,
                digital_marker=f"{COMMON_OCTAVE_PREFIX}ON",
                waveforms={
                    "I": f"{COMMON_OCTAVE_PREFIX}readout_wf",
                    "Q": f"{COMMON_OCTAVE_PREFIX}zero_wf",
                },
            ),
            f"{COMMON_OCTAVE_PREFIX}DC_offset_pulse": QuaConfigPulseDec(
                operation=QuaConfigPulseDecOperation.CONTROL,
                length=calibration_pulse_length,
                waveforms={"single": f"{COMMON_OCTAVE_PREFIX}DC_offset_wf"},
            ),
            f"{COMMON_OCTAVE_PREFIX}Analyze_pulse": QuaConfigPulseDec(
                operation=QuaConfigPulseDecOperation.MEASUREMENT,
                length=integration_length,
                waveforms={
                    "I": f"{COMMON_OCTAVE_PREFIX}zero_wf",
                    "Q": f"{COMMON_OCTAVE_PREFIX}zero_wf",
                },
                integration_weights={
                    "integW_cos": f"{COMMON_OCTAVE_PREFIX}integW_cosine",
                    "integW_sin": f"{COMMON_OCTAVE_PREFIX}integW_sine",
                    "integW_minus_sin": f"{COMMON_OCTAVE_PREFIX}integW_minus_sine",
                    "integW_zero": f"{COMMON_OCTAVE_PREFIX}integW_zero",
                },
                digital_marker=f"{COMMON_OCTAVE_PREFIX}ON",
            ),
        }
    )
    pb_config.v1_beta.waveforms.update(
        {
            f"{COMMON_OCTAVE_PREFIX}readout_wf": QuaConfigWaveformDec(
                constant=QuaConfigConstantWaveformDec(sample=calibration_amp)
            ),
            f"{COMMON_OCTAVE_PREFIX}zero_wf": QuaConfigWaveformDec(constant=QuaConfigConstantWaveformDec(sample=0.0)),
            f"{COMMON_OCTAVE_PREFIX}DC_offset_wf": QuaConfigWaveformDec(
                constant=QuaConfigConstantWaveformDec(sample=offset_amp)
            ),
        }
    )
    pb_config.v1_beta.digital_waveforms.update(
        {
            f"{COMMON_OCTAVE_PREFIX}ON": QuaConfigDigitalWaveformDec(
                samples=[QuaConfigDigitalWaveformSample(value=True, length=0)]
            ),
            f"{COMMON_OCTAVE_PREFIX}OFF": QuaConfigDigitalWaveformDec(
                samples=[QuaConfigDigitalWaveformSample(value=False, length=0)]
            ),
        }
    )
    n_samples = integration_length // 4
    pb_config.v1_beta.integration_weights.update(
        {
            f"{COMMON_OCTAVE_PREFIX}integW_cosine": QuaConfigIntegrationWeightDec(
                cosine=build_iw_sample([weight_amplitude] * n_samples),
                sine=build_iw_sample([0.0] * n_samples),
            ),
            f"{COMMON_OCTAVE_PREFIX}integW_sine": QuaConfigIntegrationWeightDec(
                cosine=build_iw_sample([0.0] * n_samples),
                sine=build_iw_sample([weight_amplitude] * n_samples),
            ),
            f"{COMMON_OCTAVE_PREFIX}integW_minus_sine": QuaConfigIntegrationWeightDec(
                cosine=build_iw_sample([0.0] * n_samples),
                sine=build_iw_sample([-weight_amplitude] * n_samples),
            ),
            f"{COMMON_OCTAVE_PREFIX}integW_zero": QuaConfigIntegrationWeightDec(
                cosine=build_iw_sample([0.0] * n_samples),
                sine=build_iw_sample([0.0] * n_samples),
            ),
        }
    )
    pb_config.v1_beta.mixers.update(
        {
            f"{COMMON_OCTAVE_PREFIX}dummy_mixer": QuaConfigMixerDec(
                correction=[
                    QuaConfigCorrectionEntry(
                        frequency=abs(int(if_frequency)),
                        frequency_double=[0.0, abs(float(if_frequency))][frequency_idx],
                        lo_frequency=int(dummy_lo_frequency),
                        lo_frequency_double=[0.0, float(dummy_lo_frequency)][frequency_idx],
                        correction=QuaConfigMatrix(1.0, 0.0, 0.0, 1.0),
                        frequency_negative=if_frequency < 0,
                    )
                    for if_frequency in (
                        dummy_if_frequency,
                        dummy_if_frequency - dummy_down_mixer_offset,
                        -dummy_down_mixer_offset,
                        -dummy_if_frequency - dummy_down_mixer_offset,
                    )
                ]
            )
        }
    )
    return pb_config


def _create_lo_to_if_mapping(lo_if_frequencies_tuple_list: List[Tuple[int, int]]) -> Dict[float, Tuple[float, ...]]:
    lo_to_if_mapping: Dict[float, Tuple[float, ...]] = defaultdict(tuple)
    for lo_freq, if_freq in lo_if_frequencies_tuple_list:
        lo_to_if_mapping[float(lo_freq)] += (if_freq,)
    return dict(lo_to_if_mapping)


@lru_cache(maxsize=None)
def _cached_get_device(
    connection_info: ConnectionDetails,
    loop_backs: Tuple[Tuple[OctaveLOSource, OctaveOutput], ...],
    octave_name: str,
) -> Octave:
    standardized_loop_backs = {input_port: output_port for input_port, output_port in loop_backs}
    return Octave(
        host=connection_info.host,
        port=connection_info.port,
        port_mapping=standardized_loop_backs,
        octave_name=octave_name,
        connection_headers=connection_info.headers,
    )


def get_device(
    connection_info: ConnectionDetails,
    loop_backs: Dict[OctaveLOSource, OctaveOutput],
    octave_name: str,
    fan: Any = None,
) -> Octave:
    client = _cached_get_device(
        connection_info,
        loop_backs=tuple(sorted(loop_backs.items(), key=lambda u: (u[0].name, u[1].name))),
        octave_name=octave_name,
    )
    if fan is not None:
        client._set_fan(fan)
    return client



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/job_manager.proto
# plugin: python-betterproto
# This file has been @generated
import warnings
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
)

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

from .. import (
    errors as _errors__,
    general_messages as _general_messages__,
)


if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


class ChangeOscillatorFrequencyRequestUpdateComponentSelection(betterproto.Enum):
    both = 0
    upconverter = 1
    downconverter = 2


@dataclass(eq=False, repr=False)
class GetElementCorrectionRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    qe_name: str = betterproto.string_field(2)
    correction: "_general_messages__.Matrix" = betterproto.message_field(3)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.is_set("correction"):
            warnings.warn(
                "GetElementCorrectionRequest.correction is deprecated",
                DeprecationWarning,
            )


@dataclass(eq=False, repr=False)
class GetElementCorrectionResponse(betterproto.Message):
    job_manager_response_header: "JobManagerResponseHeader" = betterproto.message_field(
        1
    )
    correction: "_general_messages__.Matrix" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class SetElementCorrectionRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    qe_name: str = betterproto.string_field(2)
    correction: "_general_messages__.Matrix" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class SetElementCorrectionResponse(betterproto.Message):
    job_manager_response_header: "JobManagerResponseHeader" = betterproto.message_field(
        1
    )
    correction: "_general_messages__.Matrix" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class JobManagerResponseHeader(betterproto.Message):
    success: bool = betterproto.bool_field(1)
    job_id: str = betterproto.string_field(2)
    job_manager_error_type: "_errors__.JobManagerErrorTypes" = betterproto.enum_field(
        10
    )
    job_error_details: "JobErrorDetails" = betterproto.message_field(11)


@dataclass(eq=False, repr=False)
class JobErrorDetails(betterproto.Message):
    job_operation_specific_error_type: "_errors__.JobOperationSpecificErrorTypes" = (
        betterproto.enum_field(1, group="errorType")
    )
    config_query_error_type: "_errors__.ConfigQueryErrorTypes" = betterproto.enum_field(
        2, group="errorType"
    )
    message: str = betterproto.string_field(10)


@dataclass(eq=False, repr=False)
class JobOperationSpecificError(betterproto.Message):
    type: "_errors__.JobOperationSpecificErrorTypes" = betterproto.enum_field(1)
    message: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class InsertInputStreamRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    stream_name: str = betterproto.string_field(2)
    int_stream_data: "IntStreamData" = betterproto.message_field(
        3, group="stream_data_oneof"
    )
    fixed_stream_data: "FixedStreamData" = betterproto.message_field(
        4, group="stream_data_oneof"
    )
    bool_stream_data: "BoolStreamData" = betterproto.message_field(
        5, group="stream_data_oneof"
    )


@dataclass(eq=False, repr=False)
class ChangeOscillatorFrequencyRequest(betterproto.Message):
    job_id: str = betterproto.string_field(1)
    element_name: str = betterproto.string_field(2)
    new_frequency_hz: float = betterproto.double_field(3)
    update_component: "ChangeOscillatorFrequencyRequestUpdateComponentSelection" = (
        betterproto.enum_field(4)
    )


@dataclass(eq=False, repr=False)
class ChangeOscillatorFrequencyResponse(betterproto.Message):
    job_manager_response_header: "JobManagerResponseHeader" = betterproto.message_field(
        1
    )


@dataclass(eq=False, repr=False)
class IntStreamData(betterproto.Message):
    data: List[int] = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class FixedStreamData(betterproto.Message):
    data: List[float] = betterproto.double_field(1)


@dataclass(eq=False, repr=False)
class BoolStreamData(betterproto.Message):
    data: List[bool] = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class InsertInputStreamResponse(betterproto.Message):
    job_manager_response_header: "JobManagerResponseHeader" = betterproto.message_field(
        1
    )


class JobManagerServiceStub(betterproto.ServiceStub):
    async def get_element_correction(
        self,
        get_element_correction_request: "GetElementCorrectionRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "GetElementCorrectionResponse":
        return await self._unary_unary(
            "/qm.grpc.job_manager.JobManagerService/GetElementCorrection",
            get_element_correction_request,
            GetElementCorrectionResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def set_element_correction(
        self,
        set_element_correction_request: "SetElementCorrectionRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "SetElementCorrectionResponse":
        return await self._unary_unary(
            "/qm.grpc.job_manager.JobManagerService/SetElementCorrection",
            set_element_correction_request,
            SetElementCorrectionResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def insert_input_stream(
        self,
        insert_input_stream_request: "InsertInputStreamRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "InsertInputStreamResponse":
        return await self._unary_unary(
            "/qm.grpc.job_manager.JobManagerService/InsertInputStream",
            insert_input_stream_request,
            InsertInputStreamResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def update_oscillator_frequency(
        self,
        change_oscillator_frequency_request: "ChangeOscillatorFrequencyRequest",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None
    ) -> "ChangeOscillatorFrequencyResponse":
        return await self._unary_unary(
            "/qm.grpc.job_manager.JobManagerService/UpdateOscillatorFrequency",
            change_oscillator_frequency_request,
            ChangeOscillatorFrequencyResponse,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class JobManagerServiceBase(ServiceBase):
    async def get_element_correction(
        self, get_element_correction_request: "GetElementCorrectionRequest"
    ) -> "GetElementCorrectionResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def set_element_correction(
        self, set_element_correction_request: "SetElementCorrectionRequest"
    ) -> "SetElementCorrectionResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def insert_input_stream(
        self, insert_input_stream_request: "InsertInputStreamRequest"
    ) -> "InsertInputStreamResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def update_oscillator_frequency(
        self, change_oscillator_frequency_request: "ChangeOscillatorFrequencyRequest"
    ) -> "ChangeOscillatorFrequencyResponse":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_get_element_correction(
        self,
        stream: "grpclib.server.Stream[GetElementCorrectionRequest, GetElementCorrectionResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_element_correction(request)
        await stream.send_message(response)

    async def __rpc_set_element_correction(
        self,
        stream: "grpclib.server.Stream[SetElementCorrectionRequest, SetElementCorrectionResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.set_element_correction(request)
        await stream.send_message(response)

    async def __rpc_insert_input_stream(
        self,
        stream: "grpclib.server.Stream[InsertInputStreamRequest, InsertInputStreamResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.insert_input_stream(request)
        await stream.send_message(response)

    async def __rpc_update_oscillator_frequency(
        self,
        stream: "grpclib.server.Stream[ChangeOscillatorFrequencyRequest, ChangeOscillatorFrequencyResponse]",
    ) -> None:
        request = await stream.recv_message()
        response = await self.update_oscillator_frequency(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/qm.grpc.job_manager.JobManagerService/GetElementCorrection": grpclib.const.Handler(
                self.__rpc_get_element_correction,
                grpclib.const.Cardinality.UNARY_UNARY,
                GetElementCorrectionRequest,
                GetElementCorrectionResponse,
            ),
            "/qm.grpc.job_manager.JobManagerService/SetElementCorrection": grpclib.const.Handler(
                self.__rpc_set_element_correction,
                grpclib.const.Cardinality.UNARY_UNARY,
                SetElementCorrectionRequest,
                SetElementCorrectionResponse,
            ),
            "/qm.grpc.job_manager.JobManagerService/InsertInputStream": grpclib.const.Handler(
                self.__rpc_insert_input_stream,
                grpclib.const.Cardinality.UNARY_UNARY,
                InsertInputStreamRequest,
                InsertInputStreamResponse,
            ),
            "/qm.grpc.job_manager.JobManagerService/UpdateOscillatorFrequency": grpclib.const.Handler(
                self.__rpc_update_oscillator_frequency,
                grpclib.const.Cardinality.UNARY_UNARY,
                ChangeOscillatorFrequencyRequest,
                ChangeOscillatorFrequencyResponse,
            ),
        }



from typing import Dict, Type, Tuple, Union, cast

from qm.exceptions import QmQuaException
from qm.grpc.errors import JobManagerErrorTypes, ConfigQueryErrorTypes, JobOperationSpecificErrorTypes
from qm.grpc.job_manager import (
    InsertInputStreamRequest,
    JobManagerResponseHeader,
    InsertInputStreamResponse,
    GetElementCorrectionRequest,
    SetElementCorrectionRequest,
    GetElementCorrectionResponse,
    SetElementCorrectionResponse,
)

ResponseType = Union[InsertInputStreamResponse, SetElementCorrectionResponse, GetElementCorrectionResponse]

RequestType = Union[InsertInputStreamRequest, SetElementCorrectionRequest, GetElementCorrectionRequest]


class QmApiError(QmQuaException):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "QmApiError":
        return QmApiError(0, str(response))

    def __init__(self, code: int, message: str = "") -> None:
        super().__init__(message)
        self.message = message
        self.code = code


class UnspecifiedError(QmApiError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "UnspecifiedError":
        return UnspecifiedError("Unspecified operation specific error")

    def __init__(self, message: str) -> None:
        super().__init__(0, message)


class _QmJobError(QmApiError):
    """Base class for exceptions in this module."""

    pass


class MissingJobError(_QmJobError):
    """if the job isn't recognized by the server (can happen if it never ran or if it was already deleted)"""

    def __init__(self) -> None:
        super().__init__(1000)


class InvalidJobExecutionStatusError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1001)


class InvalidOperationOnSimulatorJobError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1002)


class InvalidOperationOnRealJobError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1003)


class UnknownInputStreamError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1006)


class ConfigQueryError(QmApiError):
    pass


class MissingElementError(ConfigQueryError):
    """"""

    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "MissingElementError":
        return MissingElementError(response.job_manager_response_header.job_error_details.message)

    def __init__(self, message: str):
        super().__init__(4001, message)


class MissingDigitalInputError(ConfigQueryError):
    """"""

    @staticmethod
    def _build_from_response(request: RequestType, response: ResponseType) -> "MissingDigitalInputError":
        return MissingDigitalInputError(response.job_manager_response_header.job_error_details.message)

    def __init__(self, message: str):
        super().__init__(4002, message)


class _InvalidConfigChangeError(QmApiError):
    pass


class ElementWithSingleInputError(_InvalidConfigChangeError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "ElementWithSingleInputError":
        return ElementWithSingleInputError(
            cast(Union[SetElementCorrectionRequest, GetElementCorrectionRequest], request).qe_name
        )

    def __init__(self, element_name: str):
        super().__init__(3000)
        self.element_name = element_name


class InvalidElementCorrectionError(_InvalidConfigChangeError):
    """If the correction values are invalid"""

    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "InvalidElementCorrectionError":
        return InvalidElementCorrectionError(
            response.job_manager_response_header.job_error_details.message,
            cast(Union[SetElementCorrectionRequest, GetElementCorrectionRequest], request).qe_name,
            (
                cast(Union[SetElementCorrectionRequest, GetElementCorrectionRequest], request).correction.v00,
                cast(Union[SetElementCorrectionRequest, GetElementCorrectionRequest], request).correction.v01,
                cast(Union[SetElementCorrectionRequest, GetElementCorrectionRequest], request).correction.v10,
                cast(Union[SetElementCorrectionRequest, GetElementCorrectionRequest], request).correction.v11,
            ),
        )

    def __init__(
        self,
        message: str,
        element_name: str,
        correction: Tuple[float, float, float, float],
    ) -> None:
        super().__init__(3001, message)
        self.element_name = element_name
        self.correction = correction


class ElementWithoutIntermediateFrequencyError(_InvalidConfigChangeError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "ElementWithoutIntermediateFrequencyError":
        return ElementWithoutIntermediateFrequencyError(
            cast(Union[SetElementCorrectionRequest, GetElementCorrectionRequest], request).qe_name
        )

    def __init__(self, element_name: str):
        super().__init__(3002)
        self.element_name = element_name


class InvalidDigitalInputThresholdError(_InvalidConfigChangeError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "InvalidDigitalInputThresholdError":
        return InvalidDigitalInputThresholdError(response.job_manager_response_header.job_error_details.message)

    def __init__(self, message: str):
        super().__init__(3003)
        self.message = message


class InvalidDigitalInputDeadtimeError(_InvalidConfigChangeError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "InvalidDigitalInputDeadtimeError":
        return InvalidDigitalInputDeadtimeError(response.job_manager_response_header.job_error_details.message)

    def __init__(self, message: str):
        super().__init__(3004)
        self.message = message


class InvalidDigitalInputPolarityError(_InvalidConfigChangeError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "InvalidDigitalInputPolarityError":
        return InvalidDigitalInputPolarityError(response.job_manager_response_header.job_error_details.message)

    def __init__(self, message: str):
        super().__init__(3005)
        self.message = message


def _handle_job_manager_error(
    request: RequestType,
    response: ResponseType,
    valid_errors: Tuple[Type[QmQuaException], ...],
) -> None:
    api_response: JobManagerResponseHeader = response.job_manager_response_header
    if not api_response.success:
        error_type = api_response.job_manager_error_type

        if error_type == JobManagerErrorTypes.MissingJobError:
            raise MissingJobError()
        elif error_type == JobManagerErrorTypes.InvalidJobExecutionStatusError:
            raise InvalidJobExecutionStatusError()
        elif error_type == JobManagerErrorTypes.InvalidOperationOnSimulatorJobError:
            raise InvalidOperationOnSimulatorJobError()
        elif error_type == JobManagerErrorTypes.InvalidOperationOnRealJobError:
            raise InvalidOperationOnRealJobError()
        elif error_type == JobManagerErrorTypes.JobOperationSpecificError:
            exception_to_raise = _get_handle_job_operation_error(request, response)
            if exception_to_raise is not None and type(exception_to_raise) in valid_errors:
                raise exception_to_raise
            else:
                raise UnspecifiedError("Unspecified operation specific error")
        elif error_type == JobManagerErrorTypes.ConfigQueryError:
            exception_to_raise = _get_handle_config_query_error(request, response)
            if exception_to_raise is not None and type(exception_to_raise) in valid_errors:
                raise exception_to_raise
            else:
                raise UnspecifiedError("Unspecified operation specific error")
        elif error_type == JobManagerErrorTypes.UnknownInputStreamError:
            raise UnknownInputStreamError()
        else:
            raise UnspecifiedError("Unspecified operation error")


def _get_handle_config_query_error(request: RequestType, response: ResponseType) -> QmApiError:
    error_type = response.job_manager_response_header.job_error_details.config_query_error_type
    errors: Dict[int, Type[QmApiError]] = {
        ConfigQueryErrorTypes.MissingElementError: MissingElementError,
        ConfigQueryErrorTypes.MissingDigitalInputError: MissingDigitalInputError,
    }

    return errors.get(error_type, UnspecifiedError).build_from_response(request, response)


def _get_handle_job_operation_error(request: RequestType, response: ResponseType) -> QmApiError:
    error_type = response.job_manager_response_header.job_error_details.job_operation_specific_error_type
    errors: Dict[int, Type[QmApiError]] = {
        JobOperationSpecificErrorTypes.SingleInputElementError: ElementWithSingleInputError,
        JobOperationSpecificErrorTypes.InvalidCorrectionMatrixError: InvalidElementCorrectionError,
        JobOperationSpecificErrorTypes.ElementWithoutIntermediateFrequencyError: ElementWithoutIntermediateFrequencyError,
        JobOperationSpecificErrorTypes.InvalidDigitalInputThresholdError: InvalidDigitalInputThresholdError,
        JobOperationSpecificErrorTypes.InvalidDigitalInputDeadtimeError: InvalidDigitalInputDeadtimeError,
        JobOperationSpecificErrorTypes.InvalidDigitalInputPolarityError: InvalidDigitalInputPolarityError,
    }

    return errors.get(error_type, UnspecifiedError).build_from_response(request, response)



import logging

import betterproto

from qm.jobs.qm_job import QmJob
from qm.jobs.base_job import QmBaseJob
from qm.utils import run_until_with_timeout
from qm.exceptions import JobCancelledError, ErrorJobStateError, UnknownJobStateError
from qm.grpc.frontend import (
    JobExecutionStatus,
    JobExecutionStatusError,
    JobExecutionStatusLoading,
    JobExecutionStatusPending,
    JobExecutionStatusRunning,
    JobExecutionStatusCanceled,
    JobExecutionStatusCompleted,
)

logger = logging.getLogger(__name__)

INVALID_QUEUE_POSITION = -1


class QmPendingJob(QmBaseJob):
    """A Class describing a job in the execution queue"""

    def position_in_queue(self) -> int:
        """Returns the current position of the job in the queue, returns -1 on if the job is not pending anymore

        Returns:
            The position in the queue

        Example:
            ```python
            pending_job.position_in_queue()
            ```
        """
        status = self._job_manager.get_job_execution_status(self._id, self._machine_id)
        _, status_inst = betterproto.which_one_of(status, "status")
        if isinstance(status_inst, JobExecutionStatusPending):
            return status.pending.position_in_queue

        logger.warning(f"Job {self.id} is not pending, therefor it does not have a position in queue")
        return INVALID_QUEUE_POSITION

    def wait_for_execution(self, timeout: float = float("infinity")) -> QmJob:
        """Waits for the job to be executed (start running) or until the timeout has elapsed.
        On zero and negative timeout, the job is checked once.

        Args:
            timeout: Timeout (in seconds) for this operation

        Raises:
            TimeoutError: When timeout is elapsed

        Returns:
            The running ``QmJob``
        """

        def on_iteration() -> bool:
            status: JobExecutionStatus = self._job_manager.get_job_execution_status(self._id, self._machine_id)
            value = betterproto.which_one_of(status, "status")[1]
            if isinstance(value, (JobExecutionStatusRunning, JobExecutionStatusCompleted)):
                return True

            if isinstance(value, (JobExecutionStatusPending, JobExecutionStatusLoading)):
                return False

            if isinstance(value, JobExecutionStatusError):
                raise ErrorJobStateError(
                    f"job {self._id} encountered an error",
                    error_list=[value.string_value for value in value.error_messages.values],
                )

            elif isinstance(value, JobExecutionStatusCanceled):
                raise JobCancelledError(f"job {self._id} was cancelled")

            else:
                raise UnknownJobStateError(f"job {self._id} has unknown job state")

        def on_complete() -> QmJob:
            return QmJob(
                job_id=self._id,
                machine_id=self._machine_id,
                frontend_api=self._frontend,
                capabilities=self._capabilities,
                store=self._store,
            )

        return run_until_with_timeout(
            on_iteration_callback=on_iteration,
            on_complete_callback=on_complete,
            timeout=timeout,
            loop_interval=0.01,
            timeout_message=f"job {self._id} was not started in time. "
            f"Check the QMApp for errors and try reloading the version. "
            f"Please also inform QM about this issue. ",
        )

    def cancel(self) -> bool:
        """Removes the job from the queue

        Returns:
            true if the operation was successful

        Example:
            ```python
            pending_job.cancel()
            ```
        """
        return self._job_manager.remove_job(self._machine_id, job_id=self._id) > 0
